Package: progressify
Version: 0.2.0
Title: Progress Reporting of Common Functions via One Magic Function
Description: The progressify() function rewrites (transpiles) calls to sequential and parallel map-reduce functions such as base::lapply(), purrr::map(), foreach::foreach(), and plyr::llply() to signal progress updates. By combining this function with R's native pipe operator, you have a straightforward way to report progress on iterative computations with minimal refactoring, e.g. 'lapply(x, fcn) |> progressify()' and 'purrr::map(x, fcn) |> progressify()'. It is compatible with the parallel-processing map-reduce packages 'future.apply', 'furrr', 'crossmap', 'foreach', 'doFuture', and 'futurize'. It also supports domain-specific packages including 'boot', 'fwb', 'lme4', 'partykit', 'sandwich', and 'SimDesign', e.g. 'boot::boot(data, stat, R) |> progressify()'.
Authors@R: c(
   person("Henrik", "Bengtsson", role=c("aut", "cre", "cph"),
          email = "henrikb@braju.com",
          comment = c(ORCID = "0000-0002-7579-5165")))
License: Apache License (>= 2)
URL: https://progressify.futureverse.org,
        https://github.com/futureverse/progressify
BugReports: https://github.com/futureverse/progressify/issues
Depends: R (>= 4.1.0), progressr
Imports: methods
Suggests: base64enc, boot, commonmark, crossmap, doFuture, foreach,
        furrr, future.apply, fwb, futurize, lme4, partykit, plyr,
        purrr, sandwich, SimDesign, tools
VignetteBuilder: progressify
Encoding: UTF-8
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-05-31 19:08:32 UTC; hb
Author: Henrik Bengtsson [aut, cre, cph] (ORCID:
    <https://orcid.org/0000-0002-7579-5165>)
Maintainer: Henrik Bengtsson <henrikb@braju.com>
Repository: CRAN
Date/Publication: 2026-05-31 19:30:02 UTC

# Generated by roxygen2: do not edit by hand

S3method(print,transpiled_call)
export(progressify)
export(progressify_supported_functions)
export(progressify_supported_packages)
export(pz)
importFrom(progressr,progressor)
importFrom(utils,capture.output)
importFrom(utils,getS3method)
importFrom(utils,globalVariables)
importFrom(utils,packageVersion)
importFrom(utils,str)

<div id="badges"><!-- pkgdown markup -->
<a href="https://CRAN.R-project.org/web/checks/check_results_progressify.html"><img border="0" src="https://www.r-pkg.org/badges/version/progressify" alt="CRAN check status"/></a> <a href="https://github.com/futureverse/progressify/actions?query=workflow%3AR-CMD-check"><img border="0" src="https://github.com/futureverse/progressify/actions/workflows/R-CMD-check.yaml/badge.svg?branch=develop" alt="R CMD check status"/></a>     <a href="https://app.codecov.io/gh/futureverse/progressify"><img border="0" src="https://codecov.io/gh/futureverse/progressify/branch/develop/graph/badge.svg" alt="Coverage Status"/></a> 
</div>

# progressify: Progress Reporting of Common Functions via One Magic Function <img src="man/figures/progressify-logo.png" alt="The hexlogo for the 'progressify' package" style="width: 120px; float: right; margin-right: 1ex; margin-left: 1ex;"/>

## TL;DR 

The **progressify** package makes it extremely simple to report
progress updates for your existing map-reduce calls. All you need to
know is that there is a single function called `progressify()` that
will take care of everything, e.g.

```r
y <- lapply(x, fcn) |> progressify()
y <- map(x, fcn) |> progressify()
y <- foreach(x = xs) %do% { fcn(x) } |> progressify()
res <- boot::boot(data, statistic, R = 100) |> progressify()
```

The `progressify()` function signals progress updates via the
**[progressr]** package, meaning you can choose among the **[supported
progressr handlers]** on how to render progress updates, whether it be
via the terminal, a progress bar, or even a sound. The **progressify**
package has only one hard dependency - the **[progressr]** package.

In addition to getting progress updates via **progressr**,
by using `progressify()` you also get access to all the benefits that
come with **progressr**. Notably, you have full control over when
and how progress is reported, and it works seamlessly across
different environments and frontends.


## Supported map-reduce packages

The **progressify** package supports transpilation of functions from
multiple packages. The table below summarizes the supported map-reduce
functions. To programmatically see which packages are currently
supported, use:

```r
progressify_supported_packages()
```

To see which functions are supported for a specific package, use:

```r
progressify_supported_functions("purrr")
```

| Package            | Functions                                                                                                          |
|--------------------|--------------------------------------------------------------------------------------------------------------------|
| **base**           | `apply()`, `by()`, `eapply()`, `lapply()`, `.mapply()`, `mapply()`, `Map()`, `replicate()`, `sapply()`, `tapply()`, `vapply()` |
| **stats**          | `dendrapply()`                                                                                             |
| **[future.apply]** | `future_apply()`, `future_by()`, `future_eapply()`, `future_lapply()`, `future_.mapply()`, `future_mapply()`, `future_Map()`, `future_replicate()`, `future_sapply()`, `future_tapply()`, `future_vapply()` |
| **[purrr]**        | `map()` and variants, `walk()` and variants, `map2()` and variants, `walk2()` and variants, `pmap()` and variants, `pwalk()`, `imap()` and variants, `modify()`, `modify2()`, `imodify()` |
| **[crossmap]**     | `xmap()` and variants, `xwalk()`, `map_vec()`, `map2_vec()`, `pmap_vec()`, `imap_vec()`, plus `future_*()` variants |
| **[furrr]**        | `future_map()` and variants, `future_walk()` and variants, `future_map2()` and variants, `future_walk2()` and variants, `future_pmap()` and variants, `future_pwalk()`, `future_imap()` and variants, `future_modify()`, `future_modify2()`, `future_imodify()` and variants |
| **[foreach]**      | `%do%`, `%dopar%`                                                                                                  |
| **[doFuture]**     | `%dofuture%`                                                                                                       |
| **[plyr]**         | `llply()` and variants, `mlply()` and variants, `rdply()`, `rlply()`, `raply()`, `r_ply()` |
_Table 1: Map-reduce functions currently supported by `progressify()` for progress reporting._

Here are some examples:

```r
library(progressify)
handlers(global = TRUE)

xs <- 1:10
ys <- lapply(xs, function(x) { Sys.sleep(0.1); sqrt(x) }) |> progressify()

d <- as.dendrogram(hclust(dist(USArrests)))
d2 <- dendrapply(d, function(n) { Sys.sleep(0.01); n }) |> progressify()

xs <- 1:10
ys <- purrr::map(xs, function(x) { Sys.sleep(0.1); sqrt(x) }) |> progressify()

xs <- list(1:5, 1:5)
ys <- crossmap::xmap(xs, ~ .y * .x) |> progressify()

library(foreach)
xs <- 1:10
ys <- foreach(x = xs) %do% { Sys.sleep(0.1); sqrt(x) } |> progressify()

xs <- 1:10
ys <- plyr::llply(xs, function(x) { Sys.sleep(0.1); sqrt(x) }) |> progressify()

```


## Supported domain-specific packages

You can also progressify calls from a growing set of domain-specific
CRAN packages that have optional built-in support for
parallelization.

### CRAN packages with support for progressify

| Package                    | Functions                                                                    |
|----------------------------|------------------------------------------------------------------------------|
| **[boot]**                 | `boot()`, `censboot()`, `tsboot()`                                           |
| **[fwb]**                  | `fwb()`                                                                      |
| **[lme4]**                 | `bootMer()`                                                                  |
| **[partykit]**             | `cforest()`                                                                  |
| **[sandwich]**             | `vcovBS()`, `vcovJK()`                                                       |
| **[SimDesign]**            | `runSimulation()`                                                            |

_Table 2: CRAN packages with domain-specific functions currently supported by `progressify()` for progress reporting._

Here are some examples:

```r
res <- boot::boot(data, statistic, R = 100) |> progressify()

res <- fwb::fwb(data, statistic, R = 100) |> progressify()

res <- lme4::bootMer(fit, statistic, nsim = 100) |> progressify()

forest <- partykit::cforest(Survived ~ ., data = as.data.frame(Titanic), ntree = 50L) |> progressify()

v <- sandwich::vcovBS(fit) |> progressify()

res <- SimDesign::runSimulation(design, replications, generate, analyse, summarise) |> progressify()
```


## Compatible with futurize

The **progressify** package is compatible with the **[futurize]**
package, which parallelizes code via the [futureverse]. You can
combine the two, in either order, to get both progress reporting and
parallelization:

```r
library(progressify)
handlers(global = TRUE)

library(futurize)
plan(multisession)

xs <- 1:100
ys <- lapply(xs, slow_fcn) |> progressify() |> futurize()

ys <- purrr::map(xs, slow_fcn) |> progressify() |> futurize()

library(foreach)
ys <- foreach(x = xs) %do% { slow_fcn(x) } |> progressify() |> futurize()

ys <- plyr::llply(xs, slow_fcn) |> progressify() |> futurize()

forest <- partykit::cforest(dist ~ speed, data = cars, ntree = 50L) |> progressify() |> futurize()
```


[futureverse]: https://www.futureverse.org
[boot]: https://cran.r-project.org/package=boot
[crossmap]: https://cran.r-project.org/package=crossmap
[doFuture]: https://doFuture.futureverse.org
[foreach]: https://cran.r-project.org/package=foreach
[future.apply]: https://future.apply.futureverse.org
[futurize]: https://futurize.futureverse.org
[furrr]: https://furrr.futureverse.org
[fwb]: https://cran.r-project.org/package=fwb
[lme4]: https://cran.r-project.org/package=lme4
[partykit]: https://cran.r-project.org/package=partykit
[plyr]: https://cran.r-project.org/package=plyr
[progressr]: https://progressr.futureverse.org
[purrr]: https://cran.r-project.org/package=purrr
[sandwich]: https://cran.r-project.org/package=sandwich
[SimDesign]: https://cran.r-project.org/package=SimDesign
[supported progressr handlers]: https://progressr.futureverse.org/articles/progressr-11-handlers.html

if (requireNamespace("sandwich", quietly = TRUE)) {
  library(progressify)
  library(sandwich)
  library(stats)

  options(progressify.debug = TRUE)

  fit <- lm(dist ~ speed, data = cars)

  exprs <- list(
    vcovBS = quote(vcovBS(fit, R = 10L)),
    vcovJK = quote(vcovJK(fit))
  )
  for (name in names(exprs)) {
    message(sprintf("Testing %s ...", name))
    expr <- exprs[[name]]
    
    ## (a) Verify identical to truth
    set.seed(42)
    truth <- eval(expr)
    
    set.seed(42)
    res <- eval(bquote(.(expr) |> progressify()))
    stopifnot(all.equal(res, truth))

    ## (b) No stdout leakage
    output <- utils::capture.output({
      set.seed(42)
      res <- eval(bquote(.(expr) |> progressify()))
    })
    stopifnot(length(output) == 0L)

    ## (c) Repeated evaluation produces identical results
    set.seed(42)
    res3 <- eval(bquote(.(expr) |> progressify()))
    stopifnot(all.equal(res3, res))
    
    message(sprintf("Testing %s ... done", name))
  }
}

#' @tags pkg-partykit
if (requireNamespace("partykit")) {

library(progressify)
library(partykit)

options(progressify.debug = TRUE)

data("Titanic", package = "datasets")
tt <- as.data.frame(Titanic)

## -------------------------------------------------------
## cforest
## -------------------------------------------------------
exprs <- list(
  cforest = quote(cforest(Survived ~ ., data = tt, ntree = 20L)),
  cforest = quote(partykit::cforest(Survived ~ ., data = tt, ntree = 20L))
)

for (kk in seq_along(exprs)) {
  name <- names(exprs)[kk]
  expr <- exprs[[kk]]
  message()
  message(sprintf("=== %s ==========================", name))
  print(expr)
  message(sprintf("---------------------------------"))
  
  ## Fix random seed for reproducibility
  set.seed(1L)
  truth <- eval(expr)

  expr_f <- bquote(.(expr) |> progressify())
  print(expr_f)

  set.seed(1L)
  res <- eval(expr_f)

  ## We can't use identical() because cforest objects might contain 
  ## different call objects (e.g. including the injected applyfun)
  ## but we can check if the predicted values are identical.
  pred_truth <- predict(truth)
  pred_res <- predict(res)

  if (!identical(pred_res, pred_truth)) {
    str(list(truth = pred_truth, res = pred_res))
    stop("Predictions are not identical")
  } else {
    message("Predictions match!")
  }

  out <- utils::capture.output({
    expr_f2 <- bquote(.(expr) |> progressify())
    set.seed(1L)
    res2 <- eval(expr_f2)
  })
  print(out)
  stopifnot(identical(out, character(0L)))
  
  pred_res2 <- predict(res2)
  stopifnot(identical(pred_res2, pred_res))

  expr_f3 <- bquote(.(expr) |> progressify())
  set.seed(1L)
  res3 <- eval(expr_f3)
  pred_res3 <- predict(res3)
  stopifnot(identical(pred_res3, pred_res))
}

} # if (requireNamespace("partykit"))

#' @tags pkg-purrr
#' @tags pkg-futurize
if (requireNamespace("purrr") && requireNamespace("furrr") && requireNamespace("futurize")) {

library(progressify)
library(futurize)
library(purrr)

options(progressify.debug = TRUE)

xs <- list(aa = 1, bb = 1:2, cc = 1:10, dd = 1:5, .ee = -6:6)
FUN <- function(x, na.rm = TRUE) {
  a <- 1:5
  add <- NULL
  if (length(x) == 2) {
    add <- list(C = 42)
  }
  median(c(a, x), na.rm = na.rm)
}

ys <- list(aa = 10, bb = 20:21, cc = 30:39, dd = 40:44, .ee = 50:60)
FUN2 <- function(x, y) {
  sum(x) + sum(y)
}


exprs <- list(
  map = quote(map(xs, FUN)),
  map = quote(purrr::map(xs, FUN)),
  map = quote(purrr::map(.x = xs, .f = FUN)),

  map_lgl = quote(purrr::map_lgl(.x = xs, .f = function(x) length(x) > 1)),

  map_int = quote(purrr::map_int(.x = xs, .f = length)),

  map_dbl = quote(purrr::map_dbl(.x = xs, .f = function(x) median(c(1:5, x)))),

  map_chr = quote(purrr::map_chr(.x = xs, .f = function(x) as.character(length(x)))),

  walk = quote(purrr::walk(.x = xs, .f = function(x) invisible(NULL))),

  imap = quote(purrr::imap(xs, function(x, idx) list(idx = idx, len = length(x)))),
  imap_int = quote(purrr::imap_int(xs, function(x, idx) nchar(idx))),

  map2 = quote(purrr::map2(.x = xs, .y = ys, .f = FUN2)),
  map2 = quote(map2(xs, ys, FUN2)),
  map2_dbl = quote(purrr::map2_dbl(.x = xs, .y = ys, .f = FUN2)),

  walk2 = quote(purrr::walk2(.x = xs, .y = ys, .f = function(x, y) invisible(NULL))),

  pmap = quote(purrr::pmap(.l = list(xs, ys), .f = FUN2)),
  pmap = quote(pmap(list(xs, ys), FUN2)),
  pmap_dbl = quote(purrr::pmap_dbl(.l = list(xs, ys), .f = FUN2)),

  pwalk = quote(purrr::pwalk(.l = list(xs, ys), .f = function(x, y) invisible(NULL)))
)

for (kk in seq_along(exprs)) {
  name <- names(exprs)[kk]
  expr <- exprs[[kk]]
  message()
  message(sprintf("=== %s ==========================", name))
  print(expr)
  message(sprintf("---------------------------------"))
  truth <- eval(expr)

  for (pipe in c("progressify_futurize", "futurize_progressify")) {
    if (pipe == "progressify_futurize") {
      expr_f <- bquote(.(expr) |> progressify() |> futurize())
    } else {
      expr_f <- bquote(.(expr) |> futurize() |> progressify())
    }
    message(sprintf("--- %s ---", pipe))
    print(expr_f)

    res <- eval(expr_f)

    if (!identical(res, truth)) {
      str(list(truth = truth, res = res))
      stop("Not identical")
    } else {
      str(res)
    }

    out <- utils::capture.output({
      if (pipe == "progressify_futurize") {
        expr_f2 <- bquote(.(expr) |> progressify() |> futurize())
      } else {
        expr_f2 <- bquote(.(expr) |> futurize() |> progressify())
      }
      res2 <- eval(expr_f2)
    })
    print(out)
    stopifnot(identical(out, character(0L)))
    stopifnot(identical(res2, res))

    if (pipe == "progressify_futurize") {
      expr_f3 <- bquote(.(expr) |> progressify() |> futurize())
    } else {
      expr_f3 <- bquote(.(expr) |> futurize() |> progressify())
    }
    res3 <- eval(expr_f3)
    stopifnot(identical(res3, res))
  }
}

} # if (requireNamespace("purrr") && requireNamespace("furrr") && requireNamespace("futurize"))

library(progressify)
options(progressify.debug = TRUE)

# --------------------------------------------------------------------
# progressify()
# --------------------------------------------------------------------
message("progressify(NA):")
res <- progressify(NA)
print(res)
stopifnot(isTRUE(res))

message("progressify(FALSE):")
res <- progressify(FALSE)
print(res)
stopifnot(isTRUE(res))

message("progressify(TRUE):")
res <- progressify(TRUE)
print(res)
stopifnot(isFALSE(res))

if (requireNamespace("progressr", quietly = TRUE)) {
  message("progressify(when = FALSE):")
  y_truth <- lapply(1:3, identity)
  y <- lapply(1:3, identity) |> progressify(when = FALSE)
  stopifnot(identical(y, y_truth))
  y <- lapply(1:3, identity) |> progressify(when = TRUE)
  stopifnot(identical(y, y_truth))
  expr <- lapply(1:3, identity) |> progressify(when = FALSE, eval = FALSE)
  print(expr)
}

## Cannot progressify non-calls
res <- tryCatch(base::pi |> progressify::progressify(), error = identity)
print(res)
stopifnot(inherits(res, "error"))

## Cannot progressify non-calls
res <- tryCatch(quote(1 + 2) |> progressify::progressify(), error = identity)
print(res)
stopifnot(inherits(res, "error"))

## Cannot progressify non-existing functions
res <- tryCatch(progressify:::unknown |> progressify::progressify(), error = identity)
print(res)
stopifnot(inherits(res, "error"))

## Cannot progressify non-existing infix operators
res <- tryCatch(progressify:::`%unknown%` |> progressify::progressify(), error = identity)
print(res)
stopifnot(inherits(res, "error"))

## Cannot progressify non-supported functions
res <- tryCatch(progressify:::progressify_supported_packages() |> progressify::progressify(), error = identity)
print(res)
stopifnot(inherits(res, "error"))

## Cannot progressify private functions
res <- tryCatch(progressify:::import_progressr() |> progressify::progressify(), error = identity)
print(res)
stopifnot(inherits(res, "error"))


# --------------------------------------------------------------------
# progressify_supported_packages() and progressify_supported_functions()
# --------------------------------------------------------------------
pkgs <- progressify_supported_packages()
print(pkgs)

for (pkg in c(pkgs, "progressr", "aNonExistingPackage")) {
  cat(sprintf("Package %s:\n", pkg))
  fcns <- tryCatch({
    progressify::progressify_supported_functions(pkg)
  }, error = identity)
  print(fcns)
}

## Assert that there are not clashes between supported packages
pkgs <- progressify_supported_packages()
for (pkg in rep(pkgs, times = 2L)) {
  cat(sprintf("Package %s:\n", pkg))
  fcns <- tryCatch({
    progressify::progressify_supported_functions(pkg)
  }, error = identity)
  print(fcns)
}

#' @tags pkg-doFuture
#' @tags pkg-futurize
if (requireNamespace("foreach") && requireNamespace("doFuture") && requireNamespace("futurize")) {

library(progressify)
library(futurize)
library(foreach)

options(progressify.debug = TRUE)

xs <- 1:5
FUN <- function(x) {
  a <- 1:5
  median(c(a, x))
}


## -------------------------------------------------------
## %do%
## -------------------------------------------------------
exprs <- list(
  `%do%` = quote(
    foreach(x = xs, .combine = c) %do% { FUN(x) }
  ),

  `%do%` = quote(
    foreach::foreach(x = xs, .combine = c) %do% { FUN(x) }
  ),

  `%do%` = quote(
    foreach(x = xs) %do% { FUN(x) }
  ),

  `%do%` = quote(
    foreach(x = xs, .combine = list) %do% { FUN(x) }
  )
)

for (kk in seq_along(exprs)) {
  name <- names(exprs)[kk]
  expr <- exprs[[kk]]
  message()
  message(sprintf("=== %s ==========================", name))
  print(expr)
  message(sprintf("---------------------------------"))
  truth <- eval(expr)

  for (pipe in c("progressify_futurize", "futurize_progressify")) {
    if (pipe == "progressify_futurize") {
      expr_f <- bquote(.(expr) |> progressify() |> futurize())
    } else {
      expr_f <- bquote(.(expr) |> futurize() |> progressify())
    }
    message(sprintf("--- %s ---", pipe))
    print(expr_f)

    res <- eval(expr_f)

    if (!identical(res, truth)) {
      str(list(truth = truth, res = res))
      stop("Not identical")
    } else {
      str(res)
    }

    out <- utils::capture.output({
      if (pipe == "progressify_futurize") {
        expr_f2 <- bquote(.(expr) |> progressify() |> futurize())
      } else {
        expr_f2 <- bquote(.(expr) |> futurize() |> progressify())
      }
      res2 <- eval(expr_f2)
    })
    print(out)
    stopifnot(identical(out, character(0L)))
    stopifnot(identical(res2, res))

    if (pipe == "progressify_futurize") {
      expr_f3 <- bquote(.(expr) |> progressify() |> futurize())
    } else {
      expr_f3 <- bquote(.(expr) |> futurize() |> progressify())
    }
    res3 <- eval(expr_f3)
    stopifnot(identical(res3, res))
  }
}

} # if (requireNamespace("foreach") && requireNamespace("doFuture") && requireNamespace("futurize"))

#' @tags pkg-fwb
if (requireNamespace("fwb", quietly = TRUE)) {
  library(progressify)
  library(fwb)

  options(progressify.debug = TRUE)

  # 1. fwb::fwb test setups
  data <- mtcars
  my_stat <- function(data, w) {
    coef(lm(mpg ~ cyl, data = data, weights = w))
  }

  exprs <- list(
    fwb_simple_true = quote(fwb(data = data, statistic = my_stat, R = 10L, simple = TRUE, verbose = FALSE)),
    fwb_simple_false = quote(fwb(data = data, statistic = my_stat, R = 10L, simple = FALSE, verbose = FALSE)),
    fwb_verbose_default = quote(fwb(data = data, statistic = my_stat, R = 10L))
  )

  for (name in names(exprs)) {
    message(sprintf("Testing %s ...", name))
    expr <- exprs[[name]]

    # Ensure reproducible seed
    set.seed(42)
    truth <- eval(expr)

    set.seed(42)
    res <- eval(bquote(.(expr) |> progressify()))

    # Ensure results are equivalent
    # We compare the $t slot (simulations results matrix) and $t0 (original estimate)
    stopifnot(all.equal(res$t, truth$t))
    stopifnot(all.equal(res$t0, truth$t0))

    # Ensure no stdout leakage
    output <- utils::capture.output({
      set.seed(42)
      res2 <- eval(bquote(.(expr) |> progressify()))
    })
    stopifnot(length(output) == 0L)

    # Ensure repeated evaluation is identical
    stopifnot(all.equal(res2$t, res$t))

    message(sprintf("Testing %s ... done", name))
  }
}

#' @tags pkg-lme4
if (requireNamespace("lme4", quietly = TRUE)) {
  library(progressify)
  library(lme4)

  options(progressify.debug = TRUE)

  # 1. Fit standard linear mixed model
  fm1 <- lmer(Reaction ~ Days + (Days | Subject), sleepstudy)
  my_stat <- function(fit) {
    fixef(fit)
  }

  exprs <- list(
    bootMer = quote(bootMer(fm1, my_stat, nsim = 10L))
  )

  for (name in names(exprs)) {
    message(sprintf("Testing %s ...", name))
    expr <- exprs[[name]]

    # Ensure reproducible seed
    set.seed(42)
    truth <- eval(expr)

    set.seed(42)
    res <- eval(bquote(.(expr) |> progressify()))

    # Ensure results are equivalent
    stopifnot(all.equal(res$t, truth$t))
    stopifnot(all.equal(res$t0, truth$t0))

    # Ensure no stdout leakage
    output <- utils::capture.output({
      set.seed(42)
      res2 <- eval(bquote(.(expr) |> progressify()))
    })
    stopifnot(length(output) == 0L)

    # Ensure repeated evaluation is identical
    stopifnot(all.equal(res2$t, res$t))

    message(sprintf("Testing %s ... done", name))
  }
}

#' @tags pkg-crossmap
if (requireNamespace("crossmap")) {

library(progressify)
library(crossmap)

options(progressify.debug = TRUE)

xs <- list(1:5, 1:5)
fcn <- function(x, y) x * y


## -------------------------------------------------------
## xmap family
## -------------------------------------------------------
exprs <- list(
  xmap     = quote(xmap(xs, fcn)),
  xmap     = quote(crossmap::xmap(xs, fcn)),
  xmap     = quote(crossmap::xmap(.l = xs, .f = fcn)),
  xmap_dbl = quote(xmap_dbl(xs, ~ .y * .x)),
  xmap_dbl = quote(crossmap::xmap_dbl(xs, ~ .y * .x)),
  xmap_int = quote(xmap_int(xs, ~ .y * .x)),
  xmap_chr = quote(xmap_chr(xs, ~ paste(.y, "*", .x, "=", .y * .x))),
  xmap_lgl = quote(xmap_lgl(xs, ~ .y > .x)),
  xmap_vec = quote(xmap_vec(xs, fcn))
)

for (kk in seq_along(exprs)) {
  name <- names(exprs)[kk]
  expr <- exprs[[kk]]
  message()
  message(sprintf("=== %s ==========================", name))
  print(expr)
  message(sprintf("---------------------------------"))
  truth <- eval(expr)

  expr_f <- bquote(.(expr) |> progressify())
  print(expr_f)

  res <- eval(expr_f)

  if (!identical(res, truth)) {
    str(list(truth = truth, res = res))
    stop("Not identical")
  } else {
    str(res)
  }

  out <- utils::capture.output({
    expr_f2 <- bquote(.(expr) |> progressify())
    res2 <- eval(expr_f2)
  })
  print(out)
  stopifnot(identical(out, character(0L)))
  stopifnot(identical(res2, res))

  expr_f3 <- bquote(.(expr) |> progressify())
  res3 <- eval(expr_f3)
  stopifnot(identical(res3, res))
}


## -------------------------------------------------------
## xwalk
## -------------------------------------------------------
res_walk <- list()
truth_walk <- list()

expr <- quote(xwalk(xs, function(x, y) { list(x = x, y = y) }))
truth <- eval(expr)
res_walk_truth <- res_walk

res_walk <- list()
expr_f <- bquote(.(expr) |> progressify())
res <- eval(expr_f)
stopifnot(identical(res_walk, res_walk_truth))


## -------------------------------------------------------
## xmap_mat, xmap_arr
## -------------------------------------------------------
exprs <- list(
  xmap_mat = quote(xmap_mat(xs, fcn)),
  xmap_mat = quote(crossmap::xmap_mat(.l = xs, .f = fcn)),
  xmap_arr = quote(xmap_arr(xs, fcn))
)

for (kk in seq_along(exprs)) {
  name <- names(exprs)[kk]
  expr <- exprs[[kk]]
  message()
  message(sprintf("=== %s ==========================", name))
  print(expr)
  message(sprintf("---------------------------------"))
  truth <- eval(expr)

  expr_f <- bquote(.(expr) |> progressify())
  print(expr_f)

  res <- eval(expr_f)

  if (!identical(res, truth)) {
    str(list(truth = truth, res = res))
    stop("Not identical")
  } else {
    str(res)
  }

  out <- utils::capture.output({
    expr_f2 <- bquote(.(expr) |> progressify())
    res2 <- eval(expr_f2)
  })
  print(out)
  stopifnot(identical(out, character(0L)))
  stopifnot(identical(res2, res))

  expr_f3 <- bquote(.(expr) |> progressify())
  res3 <- eval(expr_f3)
  stopifnot(identical(res3, res))
}


## -------------------------------------------------------
## purrr-extension *_vec functions
## -------------------------------------------------------
ys <- 1:5
fcn2 <- function(x) x^2

exprs <- list(
  map_vec  = quote(map_vec(ys, fcn2)),
  map_vec  = quote(crossmap::map_vec(ys, fcn2)),
  map2_vec = quote(map2_vec(ys, ys, fcn)),
  map2_vec = quote(crossmap::map2_vec(ys, ys, fcn)),
  pmap_vec = quote(pmap_vec(list(ys, ys), fcn)),
  pmap_vec = quote(crossmap::pmap_vec(list(ys, ys), fcn)),
  imap_vec = quote(imap_vec(ys, ~ .x + .y)),
  imap_vec = quote(crossmap::imap_vec(ys, ~ .x + .y))
)

for (kk in seq_along(exprs)) {
  name <- names(exprs)[kk]
  expr <- exprs[[kk]]
  message()
  message(sprintf("=== %s ==========================", name))
  print(expr)
  message(sprintf("---------------------------------"))
  truth <- eval(expr)

  expr_f <- bquote(.(expr) |> progressify())
  print(expr_f)

  res <- eval(expr_f)

  if (!identical(res, truth)) {
    str(list(truth = truth, res = res))
    stop("Not identical")
  } else {
    str(res)
  }

  out <- utils::capture.output({
    expr_f2 <- bquote(.(expr) |> progressify())
    res2 <- eval(expr_f2)
  })
  print(out)
  stopifnot(identical(out, character(0L)))
  stopifnot(identical(res2, res))

  expr_f3 <- bquote(.(expr) |> progressify())
  res3 <- eval(expr_f3)
  stopifnot(identical(res3, res))
}

} # if (requireNamespace("crossmap"))

#' @tags pkg-plyr
if (requireNamespace("plyr")) {

library(progressify)
library(plyr)

options(progressify.debug = TRUE)

xs <- list(aa = 1, bb = 1:2, cc = 1:10, dd = 1:5, .ee = -6:6)
FUN <- function(x, na.rm = TRUE) {
  a <- 1:5
  add <- NULL
  if (length(x) == 2) add <- list(C = 42)
  median(c(a, x), na.rm = na.rm)
}


## -------------------------------------------------------
## l*ply
## -------------------------------------------------------
exprs <- list(
  llply = quote(llply(xs, FUN)),
  llply = quote(plyr::llply(xs, FUN)),
  llply = quote(plyr::llply(.data = xs, .fun = FUN)),

  ldply = quote(plyr::ldply(.data = xs, .fun = function(x) data.frame(med = median(c(1:5, x))))),

  laply = quote(plyr::laply(.data = xs, .fun = function(x) median(c(1:5, x)))),

  l_ply = quote(l_ply(xs, FUN)),
  l_ply = quote(plyr::l_ply(xs, FUN))
)

for (kk in seq_along(exprs)) {
  name <- names(exprs)[kk]
  expr <- exprs[[kk]]
  message()
  message(sprintf("=== %s ==========================", name))
  print(expr)
  message(sprintf("---------------------------------"))
  truth <- eval(expr)

  expr_f <- bquote(.(expr) |> progressify())
  print(expr_f)

  res <- eval(expr_f)

  if (!identical(res, truth)) {
    str(list(truth = truth, res = res))
    stop("Not identical")
  } else {
    str(res)
  }

  out <- utils::capture.output({
    expr_f2 <- bquote(.(expr) |> progressify())
    res2 <- eval(expr_f2)
  })
  print(out)
  stopifnot(identical(out, character(0L)))
  stopifnot(identical(res2, res))

  expr_f3 <- bquote(.(expr) |> progressify())
  res3 <- eval(expr_f3)
  stopifnot(identical(res3, res))
}


## -------------------------------------------------------
## m*ply
## -------------------------------------------------------
args_df <- data.frame(x = 1:5, y = 6:10)
FUN_m <- function(x, y) x + y

exprs <- list(
  mlply = quote(plyr::mlply(.data = args_df, .fun = FUN_m)),

  mdply = quote(plyr::mdply(.data = args_df, .fun = FUN_m)),

  maply = quote(plyr::maply(.data = args_df, .fun = FUN_m)),

  m_ply = quote(plyr::m_ply(.data = args_df, .fun = FUN_m))
)

for (kk in seq_along(exprs)) {
  name <- names(exprs)[kk]
  expr <- exprs[[kk]]
  message()
  message(sprintf("=== %s ==========================", name))
  print(expr)
  message(sprintf("---------------------------------"))
  truth <- eval(expr)

  expr_f <- bquote(.(expr) |> progressify())
  print(expr_f)

  res <- eval(expr_f)

  if (!identical(res, truth)) {
    str(list(truth = truth, res = res))
    stop("Not identical")
  } else {
    str(res)
  }

  out <- utils::capture.output({
    expr_f2 <- bquote(.(expr) |> progressify())
    res2 <- eval(expr_f2)
  })
  print(out)
  stopifnot(identical(out, character(0L)))
  stopifnot(identical(res2, res))

  expr_f3 <- bquote(.(expr) |> progressify())
  res3 <- eval(expr_f3)
  stopifnot(identical(res3, res))
}


## -------------------------------------------------------
## r*ply
## -------------------------------------------------------
exprs <- list(
  rlply = quote(plyr::rlply(.n = 5, .expr = { median(1:5) })),

  rdply = quote(plyr::rdply(.n = 5, .expr = data.frame(med = median(1:5)))),

  raply = quote(plyr::raply(.n = 5, .expr = median(1:5))),

  r_ply = quote(plyr::r_ply(.n = 5, .expr = { median(1:5) }))
)

for (kk in seq_along(exprs)) {
  name <- names(exprs)[kk]
  expr <- exprs[[kk]]
  message()
  message(sprintf("=== %s ==========================", name))
  print(expr)
  message(sprintf("---------------------------------"))
  truth <- eval(expr)

  expr_f <- bquote(.(expr) |> progressify())
  print(expr_f)

  res <- eval(expr_f)

  if (!identical(res, truth)) {
    str(list(truth = truth, res = res))
    stop("Not identical")
  } else {
    str(res)
  }

  out <- utils::capture.output({
    expr_f2 <- bquote(.(expr) |> progressify())
    res2 <- eval(expr_f2)
  })
  print(out)
  stopifnot(identical(out, character(0L)))
  stopifnot(identical(res2, res))

  expr_f3 <- bquote(.(expr) |> progressify())
  res3 <- eval(expr_f3)
  stopifnot(identical(res3, res))
}

} # if (requireNamespace("plyr"))

#' Check whether function is an S3 generic function
#'
#' @param fcn A function.
#'
#' @return
#' Return TRUE if `fcn` is an S3 generic function, otherwise FALSE.
#'
#' @noRd
is_s3_generic <- local({
  has_UseMethod <- function(expr) {
    if (is.call(expr)) {
      first <- expr[[1]]
      if (is.symbol(first) && identical(first, as.symbol("UseMethod"))) {
        return(TRUE)
      }
      for (ii in seq_along(expr)) {
        if (has_UseMethod(expr[[ii]])) return(TRUE)
      }
    }
    FALSE
  } ## has_UseMethod()

  function(fcn) {
    if (is.primitive(fcn)) return(FALSE)
    bd <- body(fcn)
    if (is.null(bd)) return(FALSE)
    has_UseMethod(bd)
  }
}) ## is_s3_generic()

#' @tags pkg-furrr
if (requireNamespace("furrr")) {

library(progressify)
library(furrr)

options(progressify.debug = TRUE)

y <- future_map(1:3, function(x) {
  print(x)
}) |> progressify()
print(y)


xs <- list(aa = 1, bb = 1:2, cc = 1:10, dd = 1:5, .ee = -6:6)
FUN <- function(x, na.rm = TRUE) {
  a <- 1:5
  add <- NULL
  if (length(x) == 2) {
    add <- list(C = 42)
  }
  median(c(a, x), na.rm = na.rm)
}

ys <- list(aa = 10, bb = 20:21, cc = 30:39, dd = 40:44, .ee = 50:60)
FUN2 <- function(x, y) {
  sum(x) + sum(y)
}


exprs <- list(
  future_map = quote(future_map(xs, FUN)),
  future_map = quote(furrr::future_map(xs, FUN)),
  future_map = quote(furrr::future_map(.x = xs, .f = FUN)),

  future_map_lgl = quote(furrr::future_map_lgl(.x = xs, .f = function(x) length(x) > 1)),

  future_map_int = quote(furrr::future_map_int(.x = xs, .f = length)),

  future_map_dbl = quote(furrr::future_map_dbl(.x = xs, .f = function(x) median(c(1:5, x)))),

  future_map_chr = quote(furrr::future_map_chr(.x = xs, .f = function(x) as.character(length(x)))),

  future_walk = quote(furrr::future_walk(.x = xs, .f = function(x) invisible(NULL))),

  future_imap = quote(furrr::future_imap(xs, function(x, idx) list(idx = idx, len = length(x)))),
  future_imap_int = quote(furrr::future_imap_int(xs, function(x, idx) nchar(idx))),

  future_map2 = quote(furrr::future_map2(.x = xs, .y = ys, .f = FUN2)),
  future_map2 = quote(future_map2(xs, ys, FUN2)),
  future_map2_dbl = quote(furrr::future_map2_dbl(.x = xs, .y = ys, .f = FUN2)),

  future_walk2 = quote(furrr::future_walk2(.x = xs, .y = ys, .f = function(x, y) invisible(NULL))),

  future_pmap = quote(furrr::future_pmap(.l = list(xs, ys), .f = FUN2)),
  future_pmap = quote(future_pmap(list(xs, ys), FUN2)),
  future_pmap_dbl = quote(furrr::future_pmap_dbl(.l = list(xs, ys), .f = FUN2)),

  future_pwalk = quote(furrr::future_pwalk(.l = list(xs, ys), .f = function(x, y) invisible(NULL))),

  ## _vec variants (furrr >= 0.4.0)
  future_map_vec  = quote(furrr::future_map_vec(.x = xs, .f = function(x) median(c(1:5, x)))),
  future_map2_vec = quote(furrr::future_map2_vec(.x = xs, .y = ys, .f = FUN2)),
  future_pmap_vec = quote(furrr::future_pmap_vec(.l = list(xs, ys), .f = FUN2)),
  future_imap_vec = quote(furrr::future_imap_vec(xs, function(x, idx) nchar(idx)))
)

for (kk in seq_along(exprs)) {
  name <- names(exprs)[kk]
  expr <- exprs[[kk]]
  message()
  message(sprintf("=== %s ==========================", name))
  print(expr)
  message(sprintf("---------------------------------"))
  truth <- eval(expr)

  expr_f <- bquote(.(expr) |> progressify())
  print(expr_f)

  res <- eval(expr_f)

  if (!identical(res, truth)) {
    str(list(truth = truth, res = res))
    stop("Not identical")
  } else {
    str(res)
  }

  out <- utils::capture.output({
    expr_f2 <- bquote(.(expr) |> progressify())
    res2 <- eval(expr_f2)
  })
  print(out)
  stopifnot(identical(out, character(0L)))
  stopifnot(identical(res2, res))

  expr_f3 <- bquote(.(expr) |> progressify())
  res3 <- eval(expr_f3)
  stopifnot(identical(res3, res))
}

} # if (requireNamespace("furrr"))

#' @tags pkg-SimDesign
if (requireNamespace("SimDesign", quietly = TRUE)) {
  library(progressify)
  library(SimDesign)

  options(progressify.debug = TRUE)

  # Create small design
  Design <- createDesign(factor1 = c(1, 2))

  Generate <- function(condition, fixed_objects = NULL) {
    rnorm(100)
  }

  Analyse <- function(condition, dat, fixed_objects = NULL) {
    mean(dat)
  }

  Summarise <- function(condition, results, fixed_objects = NULL) {
    mean(results)
  }

  exprs <- list(
    runSimulation = quote(runSimulation(design = Design, replications = 5L,
                                        generate = Generate, analyse = Analyse,
                                        summarise = Summarise, verbose = FALSE))
  )

  # Helper to strip all custom metadata attributes and runtime-varying columns
  strip_attributes <- function(df) {
    df <- as.data.frame(df)
    # Remove metadata columns that differ on each run
    for (col in c("SIM_TIME", "RAM_USED", "COMPLETED", "SEED")) {
      df[[col]] <- NULL
    }
    attributes(df) <- list(
      names = names(df),
      row.names = seq_len(nrow(df)),
      class = "data.frame"
    )
    df
  }

  for (name in names(exprs)) {
    message(sprintf("Testing %s ...", name))
    expr <- exprs[[name]]

    # Ensure reproducible seed
    set.seed(42)
    truth <- eval(expr)

    set.seed(42)
    res <- eval(bquote(.(expr) |> progressify()))

    # Ensure results are equivalent
    stopifnot(all.equal(strip_attributes(res), strip_attributes(truth)))

    # Ensure no stdout leakage
    output <- utils::capture.output({
      set.seed(42)
      res2 <- eval(bquote(.(expr) |> progressify()))
    })
    stopifnot(length(output) == 0L)

    # Ensure repeated evaluation is identical
    stopifnot(all.equal(strip_attributes(res2), strip_attributes(res)))

    message(sprintf("Testing %s ... done", name))
  }
}

#' @tags pkg-foreach
if (requireNamespace("foreach")) {

library(progressify)
library(foreach)

options(progressify.debug = TRUE)

xs <- 1:5
FUN <- function(x) {
  a <- 1:5
  median(c(a, x))
}


## -------------------------------------------------------
## %do%
## -------------------------------------------------------
exprs <- list(
  `%do%` = quote(
    foreach(x = xs, .combine = c) %do% { FUN(x) }
  ),

  `%do%` = quote(
    foreach::foreach(x = xs, .combine = c) %do% { FUN(x) }
  ),

  `%do%` = quote(
    foreach(x = xs) %do% { FUN(x) }
  ),

  `%do%` = quote(
    foreach(x = xs, .combine = list) %do% { FUN(x) }
  )
)

for (kk in seq_along(exprs)) {
  name <- names(exprs)[kk]
  expr <- exprs[[kk]]
  message()
  message(sprintf("=== %s ==========================", name))
  print(expr)
  message(sprintf("---------------------------------"))
  truth <- eval(expr)

  expr_f <- bquote(.(expr) |> progressify())
  print(expr_f)

  res <- eval(expr_f)

  if (!identical(res, truth)) {
    str(list(truth = truth, res = res))
    stop("Not identical")
  } else {
    str(res)
  }

  out <- utils::capture.output({
    expr_f2 <- bquote(.(expr) |> progressify())
    res2 <- eval(expr_f2)
  })
  print(out)
  stopifnot(identical(out, character(0L)))
  stopifnot(identical(res2, res))

  expr_f3 <- bquote(.(expr) |> progressify())
  res3 <- eval(expr_f3)
  stopifnot(identical(res3, res))
}

} # if (requireNamespace("foreach"))

library(progressify)
options(progressify.debug = TRUE)

message("*** Internals")

# --------------------------------------------------------------------
# Debug functions
# --------------------------------------------------------------------
message("debug_indent()")
oopts <- options(warn = 2L)
res <- tryCatch(progressify:::debug_indent(delta = -1L), error = identity)
print(res)
stopifnot(inherits(res, "error"))
options(oopts)


# --------------------------------------------------------------------
# .onLoad()
# --------------------------------------------------------------------
message(".onLoad()")
progressify:::.onLoad("progressify", "progressify")


# --------------------------------------------------------------------
# bquote_compile() and bquote_apply()
# --------------------------------------------------------------------
message("bquote_compile() and bquote_apply()")
bquote_compile <- progressify:::bquote_compile
bquote_apply <- progressify:::bquote_apply

tmpl <- bquote_compile(function(a = .(X), b = .(Y)) { a + b })
expr <- bquote_apply(tmpl, X = 42, Y = NULL)
f <- formals(eval(expr))
stopifnot(identical(f$a, 42))
stopifnot(is.null(f$b))


# --------------------------------------------------------------------
# import_progressr() and import_from()
# --------------------------------------------------------------------
message("import_progressr() and import_from()")
import_from <- progressify:::import_from
import_progressr <- progressify:::import_progressr

fcn <- import_progressr("progressor")
stopifnot(is.function(fcn))

fcn <- import_progressr("nonExistingFcn", default = identity)
stopifnot(identical(fcn, identity))

res <- tryCatch(
  import_progressr("nonExistingFcn"),
  error = identity
)
stopifnot(inherits(res, "error"))

fcn <- import_from("lapply", package = "base")
stopifnot(identical(fcn, base::lapply))

fcn <- import_from("nonExisting", package = "base", default = sum)
stopifnot(identical(fcn, sum))

res <- tryCatch(
  import_from("nonExisting", package = "base"),
  error = identity
)
stopifnot(inherits(res, "error"))


# --------------------------------------------------------------------
# S3 dispatching
# --------------------------------------------------------------------
message("S3 dispatch")
is_s3_generic <- progressify:::is_s3_generic
find_s3_method <- progressify:::find_s3_method

my_generic <- function(x) UseMethod("my_generic")
stopifnot(is_s3_generic(my_generic))
stopifnot(!is_s3_generic(lapply))
stopifnot(!is_s3_generic(sum))
stopifnot(!is_s3_generic(function() NULL))

my_generic.default <- function(x) "default"
my_generic.my_class <- function(x) "my_class"

obj <- structure(list(), class = "my_class")
res <- find_s3_method(my_generic, "my_generic", quote(my_generic(obj)), envir = environment())
stopifnot(identical(res$name, "my_generic.my_class"))

fempty <- function() {}
res_empty <- find_s3_method(fempty, "fempty", quote(fempty()), envir = environment())
stopifnot(is.null(res_empty))

fdots <- function(...) {}
res_dots <- find_s3_method(fdots, "fdots", quote(fdots()), envir = environment())
stopifnot(is.null(res_dots))

res_unmatched <- find_s3_method(my_generic, "my_generic", quote(my_generic(a, b, c)), envir = environment())
stopifnot(is.null(res_unmatched))

res_lit <- find_s3_method(my_generic, "my_generic", quote(my_generic(1)), envir = environment())
stopifnot(is.null(res_lit))

res_null <- find_s3_method(my_generic, "my_generic", quote(my_generic(nonexistent_obj)), envir = environment())
stopifnot(is.null(res_null))


# --------------------------------------------------------------------
# S4 dispatching
# --------------------------------------------------------------------
message("S4 dispatch")
find_s4_method <- progressify:::find_s4_method

methods::setClass("MyS4Class", slots = list(name = "character"))
methods::setGeneric("my_s4_generic", function(x) standardGeneric("my_s4_generic"))
methods::setMethod("my_s4_generic", signature = "MyS4Class", function(x) "s4_method")

obj <- methods::new("MyS4Class", name = "test")
res <- find_s4_method(my_s4_generic, "my_s4_generic", quote(my_s4_generic(obj)), envir = environment())
stopifnot(identical(res$name, "my_s4_generic"))

res_nons4 <- find_s4_method(identity, "identity", quote(identity(1)), envir = environment())
stopifnot(is.null(res_nons4))


# --------------------------------------------------------------------
# Transpiler registry
# --------------------------------------------------------------------
message("Transpiler registry")
transpilers_for_package <- progressify:::transpilers_for_package
transpiler_packages <- progressify:::transpiler_packages
list_transpilers <- progressify:::list_transpilers

transpilers_for_package(type = "test-type", package = "test-pkg", action = "reset")
transpilers_for_package(type = "test-type", package = "test-pkg", fcn = function() "test-pkg", action = "add")

res_get <- transpilers_for_package(type = "test-type", package = "test-pkg", action = "get")
stopifnot(length(res_get) == 1L)

res_list <- transpilers_for_package(type = "test-type", action = "list")
stopifnot(length(res_list) >= 1L)

res_make <- transpilers_for_package(type = "test-type", package = "test-pkg", action = "make")
stopifnot(identical(res_make, "test-pkg"))

df_pkgs <- transpiler_packages(classes = "test-type")
stopifnot(is.data.frame(df_pkgs))

df_list <- list_transpilers(class = "progressify::built-in")
stopifnot(is.data.frame(df_list))

df_filtered <- list_transpilers(pattern = "^ba", class = "progressify::built-in")
stopifnot(is.data.frame(df_filtered))


# --------------------------------------------------------------------
# parse_call() and transpile() edge cases
# --------------------------------------------------------------------
message("parse_call and transpile edge cases")
parse_call <- progressify:::parse_call
transpile <- progressify:::transpile

res <- tryCatch({
  parse_call(quote(1))
}, error = function(e) e)
stopifnot(inherits(res, "error"))

append_call_arguments <- progressify:::append_call_arguments
res_appended <- append_call_arguments(quote(my_fcn(x)), y = 2)
stopifnot(identical(res_appended, quote(my_fcn(x, y = 2))))

register_all_transpilers <- progressify:::register_all_transpilers
register_all_transpilers()

# stats::dendrapply(X = d, FUN = FUN, ...) =>
#
# local(
#   stats::dendrapply(X = {
#       .progressr_progressor <- progressr::progressor(
#         steps = 2L * attr(d, "members") - 1L)
#       d
#     }, FUN = function(..., .progressr_progressor) {
#       on.exit(.progressr_progressor())
#       FUN(...)
#     }, .progressr_progressor = .progressr_progressor)
# )
#
progressify_stats <- local({
  ## Pre-compiled bquote templates
  template_steps_dendro <- bquote_compile({
    .progressr_progressor <- progressr::progressor(
      steps = 2L * attr(.(DATA), "members") - 1L)
    .(DATA)
  })

  function(expr, fcn_name, fcn, ..., envir = parent.frame()) {
    names <- names(expr)
    if (is.null(names)) names <- rep("", length.out = length(expr))
    names <- names[-1]
    target_names <- names(formals(fcn))[seq_along(names)]
    unnamed <- setdiff(target_names, names)
    ## Handle '...' in formals - only resolve positional args before it
    ddd <- which(unnamed == "...")
    if (length(ddd) > 0L) {
      stopifnot(length(ddd) == 1L)
      unnamed <- unnamed[seq_len(ddd - 1L)]
    }
    empty_idxs <- which(names == "")
    n <- min(length(empty_idxs), length(unnamed))
    if (n > 0L) names[empty_idxs[seq_len(n)]] <- unnamed[seq_len(n)]
    names <- c("", names)

    if (fcn_name == "dendrapply") {
      idx_X <- which(names == "X")
      idx_FUN <- which(names == "FUN")

      parts <- as.list(expr)

      stopifnot(length(idx_X) == 1L)
      parts[[idx_X]] <- bquote_apply(template_steps_dendro, DATA = parts[[idx_X]])

      stopifnot(length(idx_FUN) == 1L)
      FUN <- expr[[idx_FUN]]
      parts[[idx_FUN]] <- bquote_apply(template_FUN, FUN = FUN)

      progressr_args <- list(
        ...FUN = FUN,
        .progressr_progressor = quote(.progressr_progressor)
      )
      parts <- c(parts, progressr_args)

      bquote(local(.(as.call(parts))))
    }
  } ## progressify_stats()
})


append_builtin_transpilers_for_stats <- local({
  known_fcns <- list(
    dendrapply = c
  )

  template <- bquote_compile(function(expr, options) {
    ns <- getNamespace("stats")
    fcn <- get(.(fcn_name), mode = "function", envir = ns)
    progressify_stats(expr, fcn_name = .(fcn_name), fcn = fcn, envir = parent.frame())
  })

  make_transpiler <- function(fcn_name) {
    transpiler <- eval(bquote_apply(template))
    eval(transpiler)
  }

  function() {
    ## stats::dendrapply()
    transpilers <- list()
    for (fcn_name in names(known_fcns)) {
      transpilers[[fcn_name]] <- list(
        label = sprintf("stats::%s() transpiler", fcn_name),
        transpiler = make_transpiler(fcn_name)
      )
    } ## for (fcn_name ...)
    transpilers <- list(stats = transpilers)

    append_transpilers("progressify::built-in", transpilers)

    ## Return required packages
    c("stats", "progressr")
  }
})

#' @tags pkg-stats
if (requireNamespace("stats")) {

library(progressify)

options(progressify.debug = TRUE)


## -------------------------------------------------------
## dendrapply
## -------------------------------------------------------
d <- as.dendrogram(hclust(dist(USArrests)))
FUN <- function(node) {
  attr(node, "tested") <- TRUE
  node
}

exprs <- list(
  dendrapply = quote(dendrapply(d, FUN)),
  dendrapply = quote(stats::dendrapply(d, FUN)),
  dendrapply = quote(stats::dendrapply(X = d, FUN = FUN))
)

for (kk in seq_along(exprs)) {
  name <- names(exprs)[kk]
  expr <- exprs[[kk]]
  message()
  message(sprintf("=== %s ==========================", name))
  print(expr)
  message(sprintf("---------------------------------"))
  truth <- eval(expr)

  expr_f <- bquote(.(expr) |> progressify())
  print(expr_f)

  res <- eval(expr_f)

  if (!identical(res, truth)) {
    str(list(truth = truth, res = res))
    stop("Not identical")
  } else {
    str(res)
  }

  out <- utils::capture.output({
    expr_f2 <- bquote(.(expr) |> progressify())
    res2 <- eval(expr_f2)
  })
  print(out)
  stopifnot(identical(out, character(0L)))
  stopifnot(identical(res2, res))

  expr_f3 <- bquote(.(expr) |> progressify())
  res3 <- eval(expr_f3)
  stopifnot(identical(res3, res))
}

} # if (requireNamespace("stats"))

#' @importFrom utils globalVariables
globalVariables(".")
globalVariables(c("CALL", "OPTS", "EXPR"))

#' Efficient partial substitution in expressions
#' 
#' @param expr A bquoted R expression
#' 
#' @param substitute If TRUE, `expr` is substituted.
#' 
#' @param tmpl A bquote template.
#' 
#' @param \ldots Named arguments.
#' 
#' @param envir The environment in which the bquoted values
#' should be resolved.
#' 
#' @return
#' `bquote_compile()` returns a compiled bquote template, where attribute
#'  `expression` holds the original expression `expr`.
#' 
#' `bquote_apply()` returns an expression.
#' 
#' @examples
#' tmpl <- bquote_compile(
#'   local({
#'     cl <- do.call(.(CALL), args = .(OPTS))
#'     oopts <- options(future.ClusterFuture.clusterEvalQ = "error")
#'     on.exit(options(oopts))
#'     .(EXPR)
#'   })
#' )
#' 
#' expr <- bquote_apply(tmpl, CALL = quote(foo::bar),
#'                      OPTS = list(a = 1, b = 2), EXPR = quote(1 + 2))
#' 
#' @noRd
bquote_compile <- function(expr, substitute = TRUE) {
  if (substitute) expr <- substitute(expr)
  
  tmpl <- list()
  
  unquote <- function(e, at = integer(0L)) {
    n <- length(e)
    if (n == 0L) return()

    if (is.pairlist(e)) {
      for (kk in 1:n) unquote(e[[kk]], at = c(at, kk))
      return()
    }

    if (!is.call(e)) return()
    
    ## .(<name>)?
    if (is.name(e[[1L]]) && as.character(e[[1]]) == ".") {
      ## Record location in expression tree
      entry <- list(
        expression = e[[2L]],
        at         = at
      )
      tmpl <<- c(tmpl, list(entry))
      return()
    }
  
    ## `{`, `+`, ...
    for (kk in 1:n) unquote(e[[kk]], at = c(at, kk))
  }

  dummy <- unquote(expr)
  attr(tmpl, "expression") <- expr
  tmpl
}


#' @noRd
bquote_apply <- function(tmpl, ..., envir = parent.frame()) {
  expr <- attr(tmpl, "expression")

  args <- list(...)
  if (length(args) > 0) {
    envir <- new.env(parent = envir)
    for (name in names(args)) {
      assign(name, args[[name]], envir = envir, inherits = FALSE)
    }
  }
  
  for (kk in seq_along(tmpl)) {
    entry <- tmpl[[kk]]
    value <- eval(entry[["expression"]], envir = envir)
    at <- entry[["at"]]
    
    ## Special case: Result becomes just a value
    nat <- length(at)
    if (nat == 0) return(value)

    ## Inject a NULL (needs special care) or a regular value?
    if (is.null(value)) {
      head <- if (nat == 1L) NULL else at[-nat]
      e <- if (is.null(head)) expr else expr[[head]]
      if (is.call(e)) {
        f <- as.list(e)
        f[at[nat]] <- list(NULL)
        e <- as.call(f)
      } else if (is.pairlist(e)) {
        e[at[nat]] <- list(NULL)
        e <- as.pairlist(e)
      } else {
        stop(sprintf("Unknown type of expression (please report to the maintainer): %s", sQuote(paste(deparse(e), collapse = "\\n"))))
      }
      if (is.null(head)) {
        expr <- e
      } else {
        expr[[head]] <- e
      }
    } else {
      expr[[at]] <- value
    }
  }

  expr
}

now <- function(x = Sys.time(), format = "[%H:%M:%OS3] ") {
  ## format(x, format = format) ## slower
  format(as.POSIXlt(x, tz = ""), format = format)
}

debug_indent <- local({
  prefix <- ""
  depth <- 0L
  symbols <- rep(c("|", ":", "."), times = 10L)
 
  function(delta = 0L) {
    if (delta == 0) return(prefix)
    if (delta > 0) {
      depth <<- depth + 1L
    } else if (delta < 0) {
      depth <<- depth - 1L
      if (depth < 0L) {
        calls <- paste(vapply(sys.calls(), FUN = deparse, FUN.VALUE = NA_character_), collapse = " -> ")
        warning(sprintf("[INTERNAL WARNING]: There appears to be one mdebug_pop() too many: %s", calls), call. = TRUE, immediate. = TRUE)
        depth <- 0L
      }
    }
    prefix <<- if (depth == 0) "" else paste(paste(symbols[seq_len(depth)], " "), collapse = "")
  }
})

.debug <- new.env(parent = emptyenv())
.debug$stack <- list()

mdebug_push <- function(..., debug = isTRUE(getOption("progressify.debug"))) {
  if (!debug) return()
  msg <- mdebug(..., debug = debug)
  debug_indent(+1)
  .debug$stack <- c(.debug$stack, msg)
  invisible(msg)
}

mdebugf_push <- function(..., debug = isTRUE(getOption("progressify.debug"))) {
  if (!debug) return()
  msg <- mdebugf(..., debug = debug)
  debug_indent(+1)
  .debug$stack <- c(.debug$stack, msg)
  invisible(msg)
}

mdebug_pop <- function(..., debug = isTRUE(getOption("progressify.debug"))) {
  if (!debug) return()
  n <- length(.debug$stack)
  msg <- .debug$stack[n]
  .debug$stack <- .debug$stack[-n]
  debug_indent(-1)
  mdebug(sprintf("%s done", msg), debug = debug)
}

mdebugf_pop <- function(..., debug = isTRUE(getOption("progressify.debug"))) {
  if (!debug) return()
  n <- length(.debug$stack)
  msg <- .debug$stack[n]
  .debug$stack <- .debug$stack[-n]
  debug_indent(-1)
  mdebug(sprintf("%s done", msg), debug = debug)
}

mdebug <- function(..., prefix = now(), debug = isTRUE(getOption("progressify.debug"))) {
  if (!debug) return()
  prefix <- paste(prefix, debug_indent(), sep = "")
  msg <- paste(..., sep = "")
  message(sprintf("%s%s", prefix, msg))
  invisible(msg)
}

mdebugf <- function(..., appendLF = TRUE,
                    prefix = now(), debug = isTRUE(getOption("progressify.debug"))) {
  if (!debug) return()
  prefix <- paste(prefix, debug_indent(), sep = "")
  msg <- sprintf(...)
  message(sprintf("%s%s", prefix, msg), appendLF = appendLF)
  invisible(msg)
}

#' @importFrom utils capture.output
mprint <- function(..., appendLF = TRUE, prefix = now(), debug = isTRUE(getOption("progressify.debug"))) {
  if (!debug) return()
  prefix <- paste(prefix, debug_indent(), sep = "")
  message(paste(prefix, capture.output(print(...)), sep = "", collapse = "\n"), appendLF = appendLF)
}

#' @importFrom utils capture.output str
mstr <- function(..., appendLF = TRUE, prefix = now(), debug = isTRUE(getOption("progressify.debug"))) {
  if (!debug) return()
  prefix <- paste(prefix, debug_indent(), sep = "")
  message(paste(prefix, capture.output(str(...)), sep = "", collapse = "\n"), appendLF = appendLF)
}

#' @tags pkg-base
library(progressify)

options(progressify.debug = TRUE)

y <- lapply(X = 1:3, FUN = function(x) {
  print(x)
}) |> progressify()
print(y)


xs <- list(aa = 1, bb = 1:2, cc = 1:10, dd = 1:5, .ee = -6:6)
FUN <- function(x, na.rm = TRUE) {
  a <- 1:5
  add <- NULL
  if (length(x) == 2) {
    add <- list(C = 42)
  }
  median(c(a, x), na.rm = na.rm)
}

es <- as.environment(xs)


exprs <- list(
  lapply = quote(lapply(X = xs, FUN = FUN)),
  lapply = quote(base::lapply(X = xs, FUN = FUN)),
  
  sapply = quote(sapply(X = xs, FUN = FUN)),
  sapply = quote(base::sapply(X = xs, FUN = FUN)),
  sapply = quote(base::sapply(X = xs, FUN = FUN, simplify = FALSE)),
  sapply = quote(base::sapply(X = xs, FUN = FUN, USE.NAMES = FALSE)),
  
  vapply = quote(base::vapply(X = xs, FUN.VALUE = NA_real_, FUN = FUN)),
  vapply = quote(base::vapply(
    X = xs,
    FUN.VALUE = NA_real_,
    FUN = FUN,
    USE.NAMES = FALSE
  )),
  
  eapply = quote(base::eapply(env = es, FUN = FUN)),
  eapply = quote(base::eapply(env = es, FUN = FUN, all.names = TRUE)),
  eapply = quote(base::eapply(env = es, FUN = FUN, USE.NAMES = FALSE)),
  
  replicate = quote(replicate(10, { 42 })),
  replicate = quote(replicate(n = 10, { 1 + 2 })),
  replicate = quote(base::replicate(n = 10, 3 + 4))
)

for (kk in seq_along(exprs)) {
  name <- names(exprs)[kk]
  expr <- exprs[[kk]]
  message()
  message(sprintf("=== %s ==========================", name))
  print(expr)
  message(sprintf("---------------------------------"))
  truth <- eval(expr)

  expr_f <- bquote(.(expr) |> progressify())
  print(expr_f)

  res <- eval(expr_f)

  #  res <- res[names(truth)]

  if (!identical(res, truth)) {
    str(list(truth = truth, res = res))
    stop("Not identical")
  } else {
    str(res)
  }

  out <- utils::capture.output({
    expr_f2 <- bquote(.(expr) |> progressify())
    res2 <- eval(expr_f2)
  })
  print(out)
  stopifnot(identical(out, character(0L)))
  stopifnot(identical(res2, res))

  expr_f3 <- bquote(.(expr) |> progressify())
  res3 <- eval(expr_f3)
  stopifnot(identical(res3, res))
}

.package <- new.env(parent = emptyenv())

## covr: skip=all
#' @importFrom utils packageVersion
.onLoad <- function(libname, pkgname) {
  ## R CMD build
  register_vignette_engine_during_build_only(pkgname)

  .package[["version"]] <- packageVersion(pkgname)

  update_package_option <- import_progressr("update_package_option")
  update_package_option("progressify.debug", mode = "logical")
  debug <- isTRUE(getOption("progressify.debug"))

  if (debug) {
    envs <- Sys.getenv()
    envs <- envs[grep("R_PROGRESSIFY_", names(envs), fixed = TRUE)]
    envs <- sprintf("- %s=%s", names(envs), sQuote(envs))
    mdebug(paste(c("Progressify-specific environment variables:", envs), collapse = "\n"))
  }

  register_all_transpilers()
} ## .onLoad()

#' @tags pkg-boot
if (requireNamespace("boot", quietly = TRUE)) {
  library(progressify)
  library(boot)
  if (requireNamespace("survival", quietly = TRUE)) {
    library(survival)
  }

  options(progressify.debug = TRUE)

  # 1. boot::boot test setup
  x <- 1:100
  my_stat <- function(data, i) mean(data[i])

  # 2. boot::censboot test setup
  data(aml, package = "boot")
  aml.fun <- function(data) {
    surv <- survfit(Surv(time, cens) ~ group, data = data)
    out <- NULL
    st <- 1
    for (s in seq_along(surv$strata)) {
      inds <- st:(st + surv$strata[s] - 1)
      md <- min(surv$time[inds[1 - surv$surv[inds] >= 0.5]])
      st <- st + surv$strata[s]
      out <- c(out, md)
    }
    out
  }

  # 3. boot::tsboot test setup
  lynx.fun <- function(tsb) {
    c(mean(tsb), tsb)
  }

  exprs <- list(
    boot = quote(boot(data = x, statistic = my_stat, R = 10L)),
    censboot = quote(censboot(aml, aml.fun, R = 10L, strata = aml$group)),
    tsboot = quote(tsboot(log(lynx), lynx.fun, R = 10L, l = 20, sim = "geom"))
  )

  for (name in names(exprs)) {
    message(sprintf("Testing %s ...", name))
    expr <- exprs[[name]]

    # Ensure reproducible seed
    set.seed(42)
    truth <- eval(expr)

    set.seed(42)
    res <- eval(bquote(.(expr) |> progressify()))

    # Ensure results are equivalent
    # We compare the $t slot (simulations results matrix) as calls might contain wrapped closure info
    stopifnot(all.equal(res$t, truth$t))
    stopifnot(all.equal(res$t0, truth$t0))

    # Ensure no stdout leakage
    output <- utils::capture.output({
      set.seed(42)
      res2 <- eval(bquote(.(expr) |> progressify()))
    })
    stopifnot(length(output) == 0L)

    # Ensure repeated evaluation is identical
    stopifnot(all.equal(res2$t, res$t))

    message(sprintf("Testing %s ... done", name))
  }
}

#' @tags pkg-doFuture
if (requireNamespace("foreach") && requireNamespace("doFuture")) {

library(progressify)
library(foreach)
library(doFuture)

options(progressify.debug = TRUE)

xs <- 1:5
FUN <- function(x) {
  a <- 1:5
  median(c(a, x))
}


## -------------------------------------------------------
## %dofuture%
## -------------------------------------------------------
exprs <- list(
  `%dofuture%` = quote(
    foreach(x = xs, .combine = c) %dofuture% { FUN(x) }
  ),

  `%dofuture%` = quote(
    foreach::foreach(x = xs, .combine = c) %dofuture% { FUN(x) }
  ),

  `%dofuture%` = quote(
    foreach(x = xs) %dofuture% { FUN(x) }
  ),

  `%dofuture%` = quote(
    foreach(x = xs, .combine = list) %dofuture% { FUN(x) }
  )
)

for (kk in seq_along(exprs)) {
  name <- names(exprs)[kk]
  expr <- exprs[[kk]]
  message()
  message(sprintf("=== %s ==========================", name))
  print(expr)
  message(sprintf("---------------------------------"))
  truth <- eval(expr)

  expr_f <- bquote(.(expr) |> progressify())
  print(expr_f)

  res <- eval(expr_f)

  if (!identical(res, truth)) {
    str(list(truth = truth, res = res))
    stop("Not identical")
  } else {
    str(res)
  }

  out <- utils::capture.output({
    expr_f2 <- bquote(.(expr) |> progressify())
    res2 <- eval(expr_f2)
  })
  print(out)
  stopifnot(identical(out, character(0L)))
  stopifnot(identical(res2, res))

  expr_f3 <- bquote(.(expr) |> progressify())
  res3 <- eval(expr_f3)
  stopifnot(identical(res3, res))
}

} # if (requireNamespace("foreach") && requireNamespace("doFuture"))

#! /usr/bin/env Rscript
## This runs testme test script inst/testme/test-progressify-crossmap.R
## Don't edit - it was autogenerated by inst/testme/deploy.R
progressify:::testme("progressify-crossmap")

# foreach(x = xs, .combine = c) %dofuture% { sqrt(x) } =>
#
# local({
#   .progressr_progressor <- progressr::progressor(along = xs)
#   foreach(x = xs, .combine = c) %dofuture% {
#     on.exit(.progressr_progressor())
#     sqrt(x)
#   }
# })
#
append_builtin_transpilers_for_doFuture <- local({
  known_fcns <- list(
    `%dofuture%` = c
  )

  template <- bquote_compile(function(expr, options) {
    ns <- getNamespace("doFuture")
    fcn <- get(.(fcn_name), mode = "function", envir = ns)
    progressify_foreach(expr, fcn_name = .(fcn_name), fcn = fcn, envir = parent.frame())
  })

  make_transpiler <- function(fcn_name) {
    transpiler <- eval(bquote_apply(template))
    eval(transpiler)
  }

  function() {
    ## doFuture::`%dofuture%`()
    transpilers <- list()
    for (fcn_name in names(known_fcns)) {
      transpilers[[fcn_name]] <- list(
        label = sprintf("doFuture::`%s` transpiler", fcn_name),
        transpiler = make_transpiler(fcn_name)
      )
    } ## for (fcn_name ...)
    transpilers <- list(doFuture = transpilers)

    append_transpilers("progressify::built-in", transpilers)

    ## Return required packages
    c("doFuture", "progressr")
  }
})

#' Transpile an R expression
#'
#' @param expr An \R expression, typically a function call to transpile.
#' If FALSE, then the transpiler is disabled, and if TRUE, it is re-enabled.
#' If NA, then TRUE is returned if the transpiler is enabled, otherwise FALSE.
#'
#' @param options (optional) Named options for the transpilation.
#' 
#' @param when If TRUE (default), the expression is transpiled, otherwise not.
#'
#' @param eval If TRUE (default), the transpiled expression is evaluated,
#' otherwise it is returned.
#'
#' @param envir The environment where the expression should be evaluated.
#'
#' @param type Type of the transpiler to use.
#'
#' @param unwrap (optional) A list of functions that should be considered
#' wrapping functions that the transpiler should unwrap ("enter"). This
#' allows us to transpile expressions within `{ ... }` and `local( ... )`.
#'
#' @returns
#' Returns the value of the evaluated expression `expr` if `eval = TRUE`,
#' otherwise the transpiled expression.
#' If `expr` is NA, then TRUE is returned if the transpiler is enabled,
#' otherwise FALSE.
#'
#' @keywords internal
transpile <- local({
  .enabled <- list()
  
  function(expr, options = list(...), ..., when = TRUE, eval = TRUE, envir = parent.frame(), disable = FALSE, type = "built-in", what = "transpile", unwrap = list(base::`{`, base::`(`, base::`!`, base::local, base::with, base::I, base::identity, base::invisible, base::suppressMessages, base::suppressWarnings, base::suppressPackageStartupMessages), debug = FALSE) {
    if (debug) {
      mdebug_push("transpile() ...")
      on.exit(mdebug_pop())
    }
  
    stopifnot(
      is.logical(when), length(when) == 1L, !is.na(when)
    )
  
    ## Enable or disable transpiler, or query its state?
    enabled <- .enabled[[type]]
    if (is.null(enabled)) {
      enabled <- TRUE
      .enabled[[type]] <<- enabled
    }
    
    ## e.g. transpile(TRUE), transpile(FALSE), or transpile(NA)?
    if (is.logical(expr) && length(expr) == 1L) {
      if (is.na(expr)) return(enabled)
      old_enabled <- enabled
      .enabled[[type]] <<- expr
      return(invisible(old_enabled))
    }
  
    ## Don't transpile, i.e. evaluate as-is?
    if (!enabled || !when || disable) {
      if (eval) {
        if (debug) mdebug("Evaluate call expression")
        return(eval(expr, envir = envir))
      } else {
        if (debug) mdebug("Return call expression")
        return(expr)
      }
    }

    repeat {
      ## 1a. Get a matching transpiler
      transpiler <- get_transpiler(expr, envir = envir, type = type, what = what, unwrap = unwrap, debug = debug)
    
      transpile <- transpiler[["transpiler"]]
  
      ## 1b. If not a nested transpiler function, we're done here
      if (!inherits(transpile, "transpiler")) break
  
      ## 1c. Generate transpiled expression of nested transpiler
      expr <- local({
        if (debug) {
          mdebug_push("Apply nested transpiler ...")
          on.exit({
            mprint(expr)
            mdebug_pop()
          })
          mprint(expr)
        }
        parts <- as.list(expr)
        parts$eval <- FALSE
        expr2 <- as.call(parts)
        expr <- eval(expr2, envir = envir)
        expr
      })
    }
  
  
    ## 2. Transpile
    if (debug) {
      mdebug_push("Transpile call expression ...")
    }
  
    expr_transpiled <- transpile(expr, options = options)
    class(expr_transpiled) <- c("transpiled_call", class(expr_transpiled))
    
    if (debug) {
      mprint(expr_transpiled)
      mdebug_pop()
    }
  
  
    ## 3. Evaluate or return transpiled expression?
    if (eval) {
      if (debug) mdebug("Evaluate transpiled call expression")
      eval(expr_transpiled, envir = envir)
    } else {
      if (debug) mdebug("Return transpiled call expression")
      expr_transpiled
    }
  } ## transpile()
})
class(transpile) <- c("transpiler", class(transpile))


#' Finds transpiler for S3 method for S3 generic function call and object
#'
#' @inheritParams find_s3_method
#'
#' @param type
#'
#' @return
#' A transpiler function, or NULL if none exists.
#'
#' @noRd
find_s3_method_transpiler <- function(fcn, fcn_name, call, envir, type, debug = FALSE) {
  method <- find_s3_method(fcn, fcn_name = fcn_name, call = call, envir = envir, debug = debug)
  if (is.null(method)) return(NULL)

  pkg <- method[["package"]]
  name <- method[["name"]]

  ## Look up registered transpiler for the package of the S3 method
  transpiler_sets <- get_transpilers(type)
  transpilers <- transpiler_sets[[pkg]]

  ## If non-existing, retry by first trying to register transpilers for the package
  if (is.null(transpilers)) {
    transpilers <- tryCatch({
      transpilers_for_package(type = type, package = pkg, action = "make", debug = debug)
      transpiler_sets <- get_transpilers(type)
      transpiler_sets[[pkg]]
    }, error = function(e) NULL)
  }

  ## No transpilers registered for this package?
  if (is.null(transpilers)) return(NULL)

  transpilers[[name]]
} ## find_s3_method_transpiler()


find_s4_method_transpiler <- function(fcn, fcn_name, call, envir, type, debug = FALSE) {
  method <- find_s4_method(fcn, fcn_name = fcn_name, call = call, envir = envir, debug = debug)
  if (is.null(method)) return(NULL)

  pkg <- method[["package"]]
  name <- method[["name"]]

  ## Look up registered transpiler for the package of the S4 method
  transpiler_sets <- get_transpilers(type)
  transpilers <- transpiler_sets[[pkg]]

  ## If non-existing, retry by first trying to register transpilers for the package
  if (is.null(transpilers)) {
    transpilers <- tryCatch({
      transpilers_for_package(type = type, package = pkg, action = "make", debug = debug)
      transpiler_sets <- get_transpilers(type)
      transpiler_sets[[pkg]]
    }, error = function(e) NULL)
  }

  ## No transpilers registered for this package?
  if (is.null(transpilers)) return(NULL)

  transpilers[[name]]
} ## find_s4_method_transpiler()


#' Get a registered transpiler for an R expression
#' 
#' @inheritParams transpile
#' @inheritParams parse_call
#' 
#' @param expr The R expression, which contains an R symbol or an R call
#' to be transpiled.
#'
#' @return
#' A transpiler, which is a named list with elements:
#'
#'  * `label` - a character string describing the transpiler
#'
#'  * `transpiler` - a function that takes an R expression and
#'                   an optional argument `options`
#'
#' @keywords internal
#' @noRd
get_transpiler <- function(expr, envir = parent.frame(), unwrap = list(), type, what, debug = FALSE) {
  if (debug) {
    mdebug_push("get_transpiler() ...")
    on.exit(mdebug_pop())
    mdebug_push("Finding call to be transpiled ...")
  }
  
  call_pos <- descend_wrappers(expr, envir = envir, unwrap = unwrap, what = what, debug = debug)

  call <- expr[[call_pos]]
  call_info <- parse_call(call, envir = envir, what = what, debug = debug)
  fcn <- call_info[["fcn"]]
  fcn_name <- call_info[["fcn_name"]]
  ns_name <- call_info[["ns_name"]]

  if (debug) {
    mdebugf("Position of call to be transpiled in expression: c(%s)", comma(call_pos))
    mprint(call)
    mdebug_pop()   
    mdebugf_push("Locating %s transpiler for %s::%s() of class %s ...", sQuote(type), ns_name, fcn_name, sQuote(class(fcn)[1]))
  }

  ## Special case: A nested transpiler function?
  if (inherits(fcn, "transpiler")) {
    if (debug) {
      mdebugf("Detected a nested transpiler function: %s::%s()", ns_name, fcn_name)
    }
    transpiler <- list(
      label      = fcn_name,
      transpiler = fcn
    )

    stopifnot(call_pos == 1L)
    return(transpiler)
  }

  transpiler_sets <- get_transpilers(type)
  transpilers <- transpiler_sets[[ns_name]]
  if (is.null(transpilers)) {
    if (!requireNamespace(ns_name, quietly = TRUE)) {
      info <- if (grepl("^%.*%$", fcn_name)) {
        sprintf("%s::`%s`", ns_name, fcn_name)
      } else {
        sprintf("%s::%s()", ns_name, fcn_name)
      }
      stop(sprintf("Please install %s in order to %s %s",
           sQuote(ns_name), what, info))
    }

    ## Get transpiler package addons
    ## tryCatch() is needed for cases where a package re-exports a generic
    ## from another package, e.g. scater::runPCA() is a re-export of
    ## BiocSingular::runPCA()
    req_pkgs <- tryCatch(
      transpilers_for_package(type = type, package = ns_name, action = "make", debug = debug),
      error = function(e) NULL
    )
    if (!is.null(req_pkgs)) {
      if (debug) {
        mdebugf("Required packages: [n=%d] %s", length(req_pkgs), commaq(req_pkgs))
      }

      okay <- vapply(req_pkgs, FUN.VALUE = NA, FUN = requireNamespace, quietly = TRUE)
      if (!all(okay)) {
        pkgs <- req_pkgs[!okay]
        info <- if (grepl("^%.*%$", fcn_name)) {
          sprintf("%s::`%s`", ns_name, fcn_name)
        } else {
          sprintf("%s::%s()", ns_name, fcn_name)
        }
        stop(sprintf("Please install %s in order to %s %s",
             commaq(pkgs), what, info))
      }
      transpiler_sets <- get_transpilers(type)
      transpilers <- transpiler_sets[[ns_name]]
    }
  }

  if (debug) {
    mdebugf("Namespaces registered with %s(): %s", what, commaq(names(transpiler_sets)))
  }
  
  ## Is there a registered transpiler for the function?
  if (is.null(transpilers) || ! fcn_name %in% names(transpilers)) {
    ## Fallback: S3 generic dispatching to a method in another package
    ## Note: 'call' is expr[[call_pos]] = the function head symbol, not the
    ## full call. The full call expression including arguments is needed for
    ## match.call(), so we reconstruct it here.
    full_call <- if (length(call_pos) == 1L) expr else expr[[call_pos[-length(call_pos)]]]
    if (is_s3_generic(fcn)) {
      transpiler <- find_s3_method_transpiler(fcn, fcn_name, full_call, type, envir = envir, debug = debug)
    } else if (inherits(fcn, "standardGeneric")) {
      transpiler <- find_s4_method_transpiler(fcn, fcn_name, full_call, type, envir = envir, debug = debug)
    } else {
      transpiler <- NULL
    }
    if (is.null(transpiler)) {
      if (is.null(transpilers)) {
        stop(sprintf("Function %s::%s() is not in one of the registered %s namespaces: %s", ns_name, fcn_name, what, commaq(names(transpiler_sets))))
      }
      stop(sprintf("Do not know how to %s function: %s()", what, deparse(call)))
    }
  } else {
    transpiler <- transpilers[[fcn_name]]
  }
  if (debug) {
    stopifnot(is.list(transpiler), "label" %in% names(transpiler), "transpiler" %in% names(transpiler))
    mdebugf("Transpiler description: %s", transpiler[["label"]])
    mdebug("Transpiler function:")
    mprint(transpiler[["transpiler"]])
  }

  if (length(call_pos) > 1L) {
    if (debug) {
      mdebug_push("Creating wrapper transpiler ...")
    }
    transpiler_inner <- transpiler[["transpiler"]]
    transpiler <- list(
      label      = sprintf("Apply transpiler to inner expression at c(%s)", comma(call_pos)),
      transpiler = function(expr, ...) {
        inner_pos <- call_pos[-length(call_pos)]         
        expr_inner <- expr[[inner_pos]]
        expr_inner <- transpiler_inner(expr_inner, ...)
        expr[[inner_pos]] <- expr_inner
        expr
      }
    )
    if (debug) {
      mprint(transpiler)
      mdebug_pop()
    }
  }

  if (debug) mdebugf_pop()

  transpiler
} ## get_transpiler()


.env <- new.env()
.env[["transpiler_db"]] <- list()

get_transpilers <- function(type) {
  transpiler_db <- .env[["transpiler_db"]]
  transpiler_db[[type]]
}

append_transpilers <- function(type, ...) {
  transpiler_db <- .env[["transpiler_db"]]
  transpilers <- transpiler_db[[type]]
  transpilers <- c(transpilers, ...)
  transpiler_db[[type]] <- transpilers
  .env[["transpiler_db"]] <- transpiler_db
}


list_transpilers <- function(pattern = NULL, class) {
  data <- list()
  transpiler_db <- .env[["transpiler_db"]]
  db <- transpiler_db[[class]]
  if (is.null(db)) db <- list()
  types <- names(db)
  if (!is.null(pattern)) {
    types <- grep(pattern, types, value = TRUE)
  }
  for (type in types) {
    transpilers <- db[[type]]
    fcns <- unique(names(transpilers))
    for (fcn in fcns) {
      idxs <- which(fcn == names(transpilers))
      if (length(idxs) == 1) {
        transpilers_fcn <- transpilers[idxs]
      } else {
        ## length(idxs) > 1 should not happen, but in case ...
        transpilers_fcn <- list()
        for (idx in idxs) {
          transpilers_fcn <- c(transpilers_fcn, transpilers[idx])
        }
        drop <- duplicated(names(transpilers_fcn), fromLast = TRUE)
        transpilers_fcn <- transpilers_fcn[!drop]
      }
      transpilers_fcn <- transpilers_fcn[order(names(transpilers_fcn))]
      names <- names(transpilers_fcn)
      labels <- vapply(transpilers_fcn, FUN = function(t) t$label, FUN.VALUE = "")
      dd <- data.frame(type = type, fcn = names, description = labels)
      data <- c(data, list(dd))
    }
  }
  data <- Reduce(rbind, data)
  rownames(data) <- NULL
  data
}


transpilers_for_package <- local({
  .db <- list()
  
  function(type = "default", package, fcn, action = c("add", "make", "get", "list", "reset"), debug = FALSE) {
    stopifnot(is.character(type), length(type) == 1L, !is.na(type))
    action <- match.arg(action, several.ok = FALSE)
    
    if (debug) {
      mdebugf_push("transpilers_for_package(action = %s, type = %s) ...", sQuote(action), sQuote(type))
      on.exit(mdebug_pop())
    }

    db <- .db[[type]]
    if (is.null(db)) db <- list()
    
    if (action == "add") {
      stopifnot(
        is.character(package), length(package) == 1L,
        is.function(fcn)
      )
      if (debug) {
        mdebugf(" - package: %s", sQuote(package))
      }
      fcns <- old_fcns <- db[[package]]
      fcns <- if (length(fcns) == 0) list(fcn) else c(fcns, list(fcn))
      db[[package]] <- fcns
      .db[[type]] <<- db
      invisible(old_fcns)
    } else if (action == "get") {
      if (debug) {
        mdebugf(" - package: %s", sQuote(package))
      }
      stopifnot(
        is.character(package), length(package) == 1L
      )
      if (debug) mdebugf(" - package: %s", sQuote(package))
      db[[package]]
    } else if (action == "make") {
      stopifnot(
        is.character(package), length(package) == 1L
      )
      if (debug) {
        mdebugf(" - package: %s", sQuote(package))
      }
      fcns <- db[[package]]
      if (debug) mprint(list(fcns = fcns))
      if (length(fcns) == 0L) {
        stop(sprintf("There are no factory functions for creating %s transpilers for package %s", sQuote(type), sQuote(package)))
      }
      req_pkgs <- lapply(fcns, FUN = function(fcn) fcn())
      req_pkgs <- unlist(req_pkgs, use.names = FALSE)
      req_pkgs <- sort(unique(req_pkgs))
      req_pkgs
    } else if (action == "list") {
      .db
    } else if (action == "reset") {
      old_db <- db
      db <- list()
      .db[[type]] <<- db
      invisible(old_db)
    }
  }
})


transpiler_packages <- function(classes = NULL) {
  db <- transpilers_for_package(action = "list")
  if (!is.null(classes)) {
    db <- db[names(db) %in% classes]
  }
  transpilers <- data.frame(class = character(0L), package = character(0L))
  for (class in names(db)) {
    set <- db[[class]]
    pkgs <- names(set)
    transpilers <- rbind(transpilers, data.frame(class = class, package = pkgs))
  }
  transpilers
}



#' @param package (character string) Package name.
#'
#' @param FUN A functions taking arguments `fcn` (a function),
#' `package` (character string), and `name` (character string).
#'
#' @param export (logical) If TRUE, exported functions are considered.
#'
#' @param s3methods (logical) If TRUE, registered S3 methods are considered.
#'
#' @return
#' A named list of lists of transpilers, where the names correspond
#' to function names of the package `package` and transpilers are lists.
#'
#' @noRd
make_package_transpilers <- function(package, FUN, exports = TRUE, s3methods = TRUE) {
  transpilers <- list()
  ns <- getNamespace(package)

  names <- character(0L)
  if (exports) {
    exports <- names(getNamespaceInfo(ns, "exports"))
    names <- c(names, exports)
  }
  if (s3methods) {
    s3methods <- getNamespaceInfo(ns, "S3methods")[,3]
    if (!is.character(s3methods)) {
      ## Note, although 's3methods' is typically a character vector, it
      ## might also be a list. For example, loadNamespace("strucchange")
      ## results in a character vector, but if we then load
      ## loadNamespace("partykit"), the registered S3 methods for
      ## 'strucchange' becomes a list with two function objects appended
      ## at the very end.
      s3methods <- s3methods[vapply(s3methods, FUN.VALUE = FALSE, FUN = is.character)]
      s3methods <- unlist(s3methods, use.names = FALSE)
    }
    names <- c(names, s3methods)
  }
  
  for (name in names) {
    if (exists(name, mode = "function", envir = ns, inherits = FALSE)) {
      fcn <- get(name, mode = "function", envir = ns, inherits = FALSE)
      transpilers[[name]] <- FUN(fcn, name = name)
    }
  }

  transpilers <- list(transpilers)
  names(transpilers) <- package

  transpilers
} ## make_package_transpilers()

# foreach(x = xs, .combine = c) %do% { sqrt(x) } =>
#
# local({
#   .progressr_progressor <- progressr::progressor(along = xs)
#   foreach(x = xs, .combine = c) %do% {
#     on.exit(.progressr_progressor())
#     sqrt(x)
#   }
# })
#
progressify_foreach <- local({
  ## Pre-compiled bquote templates
  template_body <- bquote_compile(local({
    on.exit(.progressr_progressor())
    .(BODY)
  }))

  function(expr, fcn_name, fcn, ..., envir = parent.frame()) {
    ## expr is:  %do%(foreach(...), { body })
    ## expr[[1]] = `%do%`
    ## expr[[2]] = foreach(...) call
    ## expr[[3]] = body expression

    foreach_call <- expr[[2]]

    ## Find the first iteration argument in the foreach() call.
    ## Iteration arguments are passed via ... and do NOT start with "."
    foreach_names <- names(foreach_call)
    if (is.null(foreach_names)) foreach_names <- rep("", length(foreach_call))
    iter_idxs <- which(nzchar(foreach_names) & !startsWith(foreach_names, "."))
    stopifnot(length(iter_idxs) >= 1L)

    ## Use the first iteration argument to determine progress steps
    iter_expr <- foreach_call[[iter_idxs[1]]]

    ## Wrap body with on.exit() progress signal
    parts <- as.list(expr)
    parts[[3]] <- bquote_apply(template_body, BODY = expr[[3]])

    ## Wrap everything in local() with progressor initialization
    bquote_apply(template_outer, ALONG = iter_expr, EXPR = as.call(parts))
  } ## progressify_foreach()
})


append_builtin_transpilers_for_foreach <- local({
  known_fcns <- list(
    `%do%` = c,
    `%dopar%` = c
  )

  template <- bquote_compile(function(expr, options) {
    ns <- getNamespace("foreach")
    fcn <- get(.(fcn_name), mode = "function", envir = ns)
    progressify_foreach(expr, fcn_name = .(fcn_name), fcn = fcn, envir = parent.frame())
  })

  make_transpiler <- function(fcn_name) {
    transpiler <- eval(bquote_apply(template))
    eval(transpiler)
  }

  function() {
    ## foreach::`%do%`(), ...
    transpilers <- list()
    for (fcn_name in names(known_fcns)) {
      transpilers[[fcn_name]] <- list(
        label = sprintf("foreach::`%s` transpiler", fcn_name),
        transpiler = make_transpiler(fcn_name)
      )
    } ## for (fcn_name ...)
    transpilers <- list(foreach = transpilers)

    append_transpilers("progressify::built-in", transpilers)

    ## Return required packages
    c("foreach", "progressr")
  }
})

# plyr::llply(.data = xs, .fun = FUN, ...) =>
#
# local(
#   plyr::llply(.data = {
#       .progressr_progressor <- progressr::progressor(along = xs)
#       xs
#     },
#     .fun = function(..., .progressr_progressor) {
#       on.exit(.progressr_progressor())
#       FUN(...)
#     }, .progressr_progressor = .progressr_progressor)
# )
#
progressify_plyr <- local({
  function(expr, fcn_name, fcn, ..., envir = parent.frame()) {
    names <- names(expr)
    if (is.null(names)) names <- rep("", length.out = length(expr))
    names <- names[-1]
    target_names <- names(formals(fcn))[seq_along(names)]
    unnamed <- setdiff(target_names, names)
    ddd <- which(unnamed == "...")
    if (length(ddd) > 0L) {
      stopifnot(length(ddd) == 1L)
      unnamed <- unnamed[seq_len(ddd - 1L)]
    }
    names[names == ""] <- unnamed
    names <- c("", names)

    idx_data <- idx_fun <- idx_n <- idx_expr <- NULL

    if (fcn_name %in% c("llply", "ldply", "laply", "l_ply")) {
      idx_data <- which(names == ".data")
      idx_fun <- which(names == ".fun")
    } else if (fcn_name %in% c("mlply", "mdply", "maply", "m_ply")) {
      idx_data <- which(names == ".data")
      idx_fun <- which(names == ".fun")
    } else if (fcn_name %in% c("rlply", "rdply", "raply", "r_ply")) {
      idx_n <- which(names == ".n")
      idx_expr <- which(names == ".expr")
    }

    parts <- as.list(expr)

    if (!is.null(idx_data)) {
      stopifnot(length(idx_data) == 1L)
      if (fcn_name %in% c("mlply", "mdply", "maply", "m_ply")) {
        parts[[idx_data]] <- bquote_apply(template_steps_nrow, DATA = parts[[idx_data]])
      } else {
        parts[[idx_data]] <- bquote_apply(template_along, ALONG = parts[[idx_data]])
      }
    }

    if (!is.null(idx_fun)) {
      stopifnot(length(idx_fun) == 1L)
      FUN <- expr[[idx_fun]]
      parts[[idx_fun]] <- bquote_apply(template_FUN, FUN = FUN)

      progressr_args <- list(
        ...FUN = FUN,
        .progressr_progressor = quote(.progressr_progressor)
      )
      parts <- c(parts, progressr_args)
    }

    if (!is.null(idx_n)) {
      stopifnot(length(idx_n) == 1L)
      parts[[idx_n]] <- bquote_apply(template_steps, STEPS = parts[[idx_n]])
    }

    if (!is.null(idx_expr)) {
      stopifnot(length(idx_expr) == 1L)
      parts[[idx_expr]] <- bquote_apply(template_expr, EXPR = expr[[idx_expr]])
    }

    bquote(local(.(as.call(parts))))
  } ## progressify_plyr()
})


append_builtin_transpilers_for_plyr <- local({
  known_fcns <- list(
    llply = c,
    ldply = c,
    laply = c,
    l_ply = c,
    mlply = c,
    mdply = c,
    maply = c,
    m_ply = c,
    rlply = c,
    rdply = c,
    raply = c,
    r_ply = c
  )

  template <- bquote_compile(function(expr, options) {
    ns <- getNamespace("plyr")
    fcn <- get(.(fcn_name), mode = "function", envir = ns)
    progressify_plyr(expr, fcn_name = .(fcn_name), fcn = fcn, envir = parent.frame())
  })

  make_transpiler <- function(fcn_name) {
    transpiler <- eval(bquote_apply(template))
    eval(transpiler)
  }

  function() {
    ## plyr::llply(), ...
    transpilers <- list()
    for (fcn_name in names(known_fcns)) {
      transpilers[[fcn_name]] <- list(
        label = sprintf("plyr::%s() transpiler", fcn_name),
        transpiler = make_transpiler(fcn_name)
      )
    } ## for (fcn_name ...)
    transpilers <- list(plyr = transpilers)

    append_transpilers("progressify::built-in", transpilers)

    ## Return required packages
    c("plyr", "progressr")
  }
})

# boot::boot(...) =>
#
# local({
#   .progressr_steps <- R
#   .progressr_first <- TRUE
#   .progressr_progressor <- progressr::progressor(steps = .progressr_steps)
#   boot::boot(
#     data = x,
#     statistic = function(..., ...FUN, .progressr_progressor) {
#       if (.progressr_first) {
#         .progressr_first <<- FALSE
#       } else {
#         on.exit(.progressr_progressor())
#       }
#       ...FUN(...)
#     },
#     R = .progressr_steps,
#     ...FUN = statistic,
#     .progressr_progressor = .progressr_progressor
#   )
# })
#
progressify_boot <- local({
  template_boot_outer <- bquote_compile(local({
    .progressr_steps <- .(STEPS)
    .progressr_first <- TRUE
    .progressr_progressor <- progressr::progressor(steps = .progressr_steps)
    .(EXPR)
  }))

  template_boot_FUN <- bquote_compile(function(..., ...FUN, .progressr_progressor) {
    if (.progressr_first) {
      .progressr_first <<- FALSE
    } else {
      on.exit(.progressr_progressor())
    }
    ...FUN(...)
  })

  function(expr, fcn_name, fcn, ..., envir = parent.frame()) {
    names <- names(expr)
    if (is.null(names)) names <- rep("", length.out = length(expr))
    names <- names[-1]
    target_names <- names(formals(fcn))[seq_along(names)]
    unnamed <- setdiff(target_names, names)
    ## Handle '...' in formals - only resolve positional args before it
    ddd <- which(unnamed == "...")
    if (length(ddd) > 0L) {
      stopifnot(length(ddd) == 1L)
      unnamed <- unnamed[seq_len(ddd - 1L)]
    }
    empty_idxs <- which(names == "")
    n <- min(length(empty_idxs), length(unnamed))
    if (n > 0L) names[empty_idxs[seq_len(n)]] <- unnamed[seq_len(n)]
    names <- c("", names)

    idx_R <- which(names == "R")
    idx_statistic <- which(names == "statistic")

    parts <- as.list(expr)

    stopifnot(length(idx_R) == 1L)
    steps <- parts[[idx_R]]
    parts[[idx_R]] <- quote(.progressr_steps)

    stopifnot(length(idx_statistic) == 1L)
    statistic <- expr[[idx_statistic]]
    parts[[idx_statistic]] <- bquote_apply(template_boot_FUN, FUN = statistic)

    progressr_args <- list(
      ...FUN = statistic,
      .progressr_progressor = quote(.progressr_progressor)
    )
    parts <- c(parts, progressr_args)

    bquote_apply(template_boot_outer, STEPS = steps, EXPR = as.call(parts))
  } ## progressify_boot()
})


append_builtin_transpilers_for_boot <- local({
  known_fcns <- list(
    boot = c,
    censboot = c,
    tsboot = c
  )

  template <- bquote_compile(function(expr, options) {
    ns <- getNamespace("boot")
    fcn <- get(.(fcn_name), mode = "function", envir = ns)
    progressify_boot(expr, fcn_name = .(fcn_name), fcn = fcn, envir = parent.frame())
  })

  make_transpiler <- function(fcn_name) {
    transpiler <- eval(bquote_apply(template))
    eval(transpiler)
  }

  function() {
    ## boot::boot(), boot::censboot(), boot::tsboot()
    transpilers <- list()
    for (fcn_name in names(known_fcns)) {
      transpilers[[fcn_name]] <- list(
        label = sprintf("boot::%s() transpiler", fcn_name),
        transpiler = make_transpiler(fcn_name)
      )
    } ## for (fcn_name ...)
    transpilers <- list(boot = transpilers)

    append_transpilers("progressify::built-in", transpilers)

    ## Return required packages
    c("boot", "progressr")
  }
})

# furrr::future_map(.x = xs, .f = FUN, ...) =>
#
# local(
#   furrr::future_map(.x = { .progressr_progressor <- progressr::progressor(along = xs); xs },
#     .f = local({
#       .progressr_f <- purrr::as_mapper(FUN)
#       function(..., .progressr_progressor) {
#         on.exit(.progressr_progressor())
#         .progressr_f(...)
#       }
#     }), .progressr_progressor = .progressr_progressor)
# )
#
progressify_furrr <- local({
  function(expr, fcn_name, fcn, ..., envir = parent.frame()) {
    names <- names(expr)
    if (is.null(names)) names <- rep("", length.out = length(expr))
    names <- names[-1]
    target_names <- names(formals(fcn))[seq_along(names)]
    unnamed <- setdiff(target_names, names)
    ddd <- which(unnamed == "...")
    if (length(ddd) > 0L) {
      stopifnot(length(ddd) == 1L)
      unnamed <- unnamed[seq_len(ddd - 1L)]
    }
    names[names == ""] <- unnamed
    names <- c("", names)

    idx_x <- idx_l <- idx_f <- NULL

    if (fcn_name %in% c("future_pmap", "future_pmap_lgl", "future_pmap_int",
                         "future_pmap_dbl", "future_pmap_chr", "future_pmap_raw",
                         "future_pmap_dfr", "future_pmap_dfc",
                         "future_pmap_vec",
                         "future_pwalk")) {
      idx_l <- which(names == ".l")
      idx_f <- which(names == ".f")
    } else if (fcn_name %in% c("future_map2", "future_map2_lgl",
                                "future_map2_int", "future_map2_dbl",
                                "future_map2_chr", "future_map2_raw",
                                "future_map2_dfr", "future_map2_dfc",
                                "future_map2_vec",
                                "future_walk2")) {
      idx_x <- which(names == ".x")
      idx_f <- which(names == ".f")
    } else {
      idx_x <- which(names == ".x")
      idx_f <- which(names == ".f")
    }

    parts <- as.list(expr)

    if (!is.null(idx_x)) {
      stopifnot(length(idx_x) == 1L)
      parts[[idx_x]] <- bquote_apply(template_along, ALONG = parts[[idx_x]])
    }

    if (!is.null(idx_l)) {
      stopifnot(length(idx_l) == 1L)
      parts[[idx_l]] <- bquote_apply(template_along_first, ALONG = parts[[idx_l]])
    }

    if (!is.null(idx_f)) {
      stopifnot(length(idx_f) == 1L)
      FUN <- expr[[idx_f]]
      ## NOTE: furrr's .f can be a function, formula, string, or integer,
      ## so we use purrr::as_mapper() to convert it to a callable function
      parts[[idx_f]] <- bquote_apply(template_f, FUN = FUN)

      progressr_args <- list(
        .progressr_progressor = quote(.progressr_progressor)
      )
      parts <- c(parts, progressr_args)
    }

    bquote(local(.(as.call(parts))))
  } ## progressify_furrr()
})


append_builtin_transpilers_for_furrr <- local({
  known_fcns <- list(
    future_map = c,
    future_map_lgl = c,
    future_map_int = c,
    future_map_dbl = c,
    future_map_chr = c,
    future_map_raw = c,
    future_map_dfr = c,
    future_map_dfc = c,
    future_map_vec = c,
    future_map_at = c,
    future_map_if = c,
    future_walk = c,
    future_map2 = c,
    future_map2_lgl = c,
    future_map2_int = c,
    future_map2_dbl = c,
    future_map2_chr = c,
    future_map2_raw = c,
    future_map2_dfr = c,
    future_map2_dfc = c,
    future_map2_vec = c,
    future_walk2 = c,
    future_pmap = c,
    future_pmap_lgl = c,
    future_pmap_int = c,
    future_pmap_dbl = c,
    future_pmap_chr = c,
    future_pmap_raw = c,
    future_pmap_dfr = c,
    future_pmap_dfc = c,
    future_pmap_vec = c,
    future_pwalk = c,
    future_imap = c,
    future_imap_lgl = c,
    future_imap_int = c,
    future_imap_dbl = c,
    future_imap_chr = c,
    future_imap_raw = c,
    future_imap_dfr = c,
    future_imap_dfc = c,
    future_imap_vec = c,
    future_iwalk = c,
    future_modify = c,
    future_modify_at = c,
    future_modify_if = c
  )

  template <- bquote_compile(function(expr, options) {
    ns <- getNamespace("furrr")
    fcn <- get(.(fcn_name), mode = "function", envir = ns)
    progressify_furrr(expr, fcn_name = .(fcn_name), fcn = fcn, envir = parent.frame())
  })

  make_transpiler <- function(fcn_name) {
    transpiler <- eval(bquote_apply(template))
    eval(transpiler)
  }

  function() {
    ## furrr::future_map(), ...
    transpilers <- list()
    for (fcn_name in names(known_fcns)) {
      transpilers[[fcn_name]] <- list(
        label = sprintf("furrr::%s() transpiler", fcn_name),
        transpiler = make_transpiler(fcn_name)
      )
    } ## for (fcn_name ...)
    transpilers <- list(furrr = transpilers)

    append_transpilers("progressify::built-in", transpilers)

    ## Return required packages
    c("furrr", "progressr")
  }
})

# lme4::bootMer(...) =>
#
# local({
#   .progressr_steps <- nsim
#   .progressr_first <- TRUE
#   .progressr_progressor <- progressr::progressor(steps = .progressr_steps)
#   .progressr_FUN <- FUN
#   lme4::bootMer(
#     x = x,
#     FUN = function(...) {
#       if (.progressr_first) {
#         .progressr_first <<- FALSE
#       } else {
#         on.exit(.progressr_progressor())
#       }
#       .progressr_FUN(...)
#     },
#     nsim = .progressr_steps
#   )
# })
#
progressify_lme4 <- local({
  template_bootMer_outer <- bquote_compile(local({
    .progressr_steps <- .(STEPS)
    .progressr_first <- TRUE
    .progressr_progressor <- progressr::progressor(steps = .progressr_steps)
    .progressr_FUN <- .(FUN)
    .(EXPR)
  }))

  template_bootMer_FUN <- quote(function(...) {
    if (.progressr_first) {
      .progressr_first <<- FALSE
    } else {
      on.exit(.progressr_progressor())
    }
    .progressr_FUN(...)
  })

  function(expr, fcn_name, fcn, ..., envir = parent.frame()) {
    names <- names(expr)
    if (is.null(names)) names <- rep("", length.out = length(expr))
    names <- names[-1]
    target_names <- names(formals(fcn))[seq_along(names)]
    unnamed <- setdiff(target_names, names)
    ## Handle '...' in formals - only resolve positional args before it
    ddd <- which(unnamed == "...")
    if (length(ddd) > 0L) {
      stopifnot(length(ddd) == 1L)
      unnamed <- unnamed[seq_len(ddd - 1L)]
    }
    empty_idxs <- which(names == "")
    n <- min(length(empty_idxs), length(unnamed))
    if (n > 0L) names[empty_idxs[seq_len(n)]] <- unnamed[seq_len(n)]
    names <- c("", names)

    idx_nsim <- which(names == "nsim")
    idx_FUN <- which(names == "FUN")

    parts <- as.list(expr)

    if (length(idx_nsim) == 1L) {
      steps <- parts[[idx_nsim]]
      parts[[idx_nsim]] <- quote(.progressr_steps)
    } else {
      steps <- 1L
      parts$nsim <- quote(.progressr_steps)
    }

    stopifnot(length(idx_FUN) == 1L)
    FUN <- expr[[idx_FUN]]
    parts[[idx_FUN]] <- template_bootMer_FUN

    bquote_apply(template_bootMer_outer, STEPS = steps, FUN = FUN, EXPR = as.call(parts))
  } ## progressify_lme4()
})


append_builtin_transpilers_for_lme4 <- local({
  known_fcns <- list(
    bootMer = c
  )

  template <- bquote_compile(function(expr, options) {
    ns <- getNamespace("lme4")
    fcn <- get(.(fcn_name), mode = "function", envir = ns)
    progressify_lme4(expr, fcn_name = .(fcn_name), fcn = fcn, envir = parent.frame())
  })

  make_transpiler <- function(fcn_name) {
    transpiler <- eval(bquote_apply(template))
    eval(transpiler)
  }

  function() {
    ## lme4::bootMer()
    transpilers <- list()
    for (fcn_name in names(known_fcns)) {
      transpilers[[fcn_name]] <- list(
        label = sprintf("lme4::%s() transpiler", fcn_name),
        transpiler = make_transpiler(fcn_name)
      )
    } ## for (fcn_name ...)
    transpilers <- list(lme4 = transpilers)

    append_transpilers("progressify::built-in", transpilers)

    ## Return required packages
    c("lme4", "progressr")
  }
})

#' List packages and functions supporting progressification
#'
#' @param package A package name.
#'
#' @return
#' A character vector of package or function names.
#'
#' @examples
#' pkgs <- progressify_supported_packages()
#' pkgs
#'
#' fcns <- progressify_supported_functions("base")
#' print(fcns)
#'
#' if (requireNamespace("purrr")) {
#'   fcns <- progressify_supported_functions("purrr")
#'   print(fcns)
#' }
#'
#' @export
progressify_supported_packages <- function() {
  transpilers <- get_transpilers("progressify::built-in")
  sort(names(transpilers))
}


#' @rdname progressify_supported_packages
#' @export
progressify_supported_functions <- function(package) {
  stopifnot(is.character(package), length(package) == 1L, !is.na(package), nzchar(package))

  transpilers <- get_transpilers("progressify::built-in")
  pkg_transpilers <- transpilers[[package]]

  if (is.null(pkg_transpilers) || length(pkg_transpilers) == 0L) {
    stop(sprintf("Package %s does not support progressification", sQuote(package)))
  }

  sort(unique(names(pkg_transpilers)))
}

## This runs 'testme' test inst/testme/test-<name>.R scripts
## Don't edit - it was autogenerated by inst/testme/deploy.R
testme <- function(name) {
  path <- system.file(package = 'progressify', 'testme', mustWork = TRUE)
  Sys.setenv(R_TESTME_PATH = path)
  Sys.setenv(R_TESTME_PACKAGE = 'progressify')
  Sys.setenv(R_TESTME_NAME = name)
  on.exit(Sys.unsetenv('R_TESTME_NAME'))
  source(file.path(path, 'run.R'))
}

## Pre-compiled bquote templates shared across transpilers

## Length-based step count from 'data' argument
template_along <- bquote_compile({
  .progressr_progressor <- progressr::progressor(along = .(ALONG))
  .(ALONG)
})

## Length-based step count from first element of list
template_along_first <- bquote_compile({
  .progressr_progressor <- progressr::progressor(along = .(ALONG)[[1]])
  .(ALONG)
})

## Step count from numeric argument
template_steps <- bquote_compile({
  .progressr_progressor <- progressr::progressor(steps = .(STEPS))
  .(STEPS)
})

## Step count from nrow() of 'data' argument
template_steps_nrow <- bquote_compile({
  .progressr_progressor <- progressr::progressor(steps = nrow(.(DATA)))
  .(DATA)
})

## Pass progressor via '...' arguments
template_FUN <- bquote_compile(function(..., ...FUN, .progressr_progressor) {
  on.exit(.progressr_progressor())
  ...FUN(...)
})

## purrr-style .f wrapper: uses as_mapper() for formula/string/integer support
template_f <- bquote_compile(local({
  .progressr_f <- purrr::as_mapper(.(FUN))
  function(..., .progressr_progressor) {
    on.exit(.progressr_progressor())
    .progressr_f(...)
  }
}))

## Wrap 'expr' with on.exit() progress signaling
template_expr <- bquote_compile(local({
  on.exit(.progressr_progressor())
  .(EXPR)
}))

## Wrap call with progressor in enclosing environment
template_outer <- bquote_compile(local({
  .progressr_progressor <- progressr::progressor(along = .(ALONG))
  .(EXPR)
}))

import_from <- function(name, default = NULL, package) {
  ns <- getNamespace(package)
  if (exists(name, mode = "function", envir = ns, inherits = FALSE)) {
    get(name, mode = "function", envir = ns, inherits = FALSE)
  } else if (!is.null(default)) {
    default
  } else {
    stop(sprintf("No such '%s' function: %s()", package, name))
  }
}

import_progressr <- function(name, default = NULL) {
  import_from(name, default = default, package = "progressr")
}

# partykit::cforest(...) =>
#
# partykit::cforest(..., applyfun = function(...) {
#   lapply(...) |> progressify::progressify()
# })
#
progressify_partykit <- local({
  function(expr, fcn_name, fcn, ..., envir = parent.frame()) {
    parts <- as.list(expr)
    parts$applyfun <- quote(function(X, FUN, ...) {
      lapply(X = X, FUN = FUN, ...) |> progressify::progressify()
    })
    as.call(parts)
  }
})


append_builtin_transpilers_for_partykit <- local({
  known_fcns <- list(
    cforest = c
  )

  template <- bquote_compile(function(expr, options) {
    ns <- getNamespace("partykit")
    fcn <- get(.(fcn_name), mode = "function", envir = ns)
    progressify_partykit(expr, fcn_name = .(fcn_name), fcn = fcn, envir = parent.frame())
  })

  make_transpiler <- function(fcn_name) {
    transpiler <- eval(bquote_apply(template))
    eval(transpiler)
  }

  function() {
    ## partykit::cforest(), ...
    transpilers <- list()
    for (fcn_name in names(known_fcns)) {
      transpilers[[fcn_name]] <- list(
        label = sprintf("partykit::%s() transpiler", fcn_name),
        transpiler = make_transpiler(fcn_name)
      )
    } ## for (fcn_name ...)
    transpilers <- list(partykit = transpilers)

    append_transpilers("progressify::built-in", transpilers)

    ## Return required packages
    c("partykit", "progressr")
  }
})

# crossmap::xmap(.l = xs, .f = FUN, ...) =>
#
# local(
#   crossmap::xmap(.l = {
#       .progressr_progressor <- progressr::progressor(
#         steps = prod(lengths(xs)))
#       xs
#     },
#     .f = local({
#       .progressr_f <- purrr::as_mapper(FUN)
#       function(..., .progressr_progressor) {
#         on.exit(.progressr_progressor())
#         .progressr_f(...)
#       }
#     }), .progressr_progressor = .progressr_progressor)
# )
#
progressify_crossmap <- local({
  ## xmap-style: number of iterations = product of all element lengths
  template_steps_prod_lengths <- bquote_compile({
    .progressr_progressor <- progressr::progressor(
      steps = prod(lengths(.(DATA))))
    .(DATA)
  })

  function(expr, fcn_name, fcn, ..., envir = parent.frame()) {
    names <- names(expr)
    if (is.null(names)) names <- rep("", length.out = length(expr))
    names <- names[-1]
    target_names <- names(formals(fcn))[seq_along(names)]
    unnamed <- setdiff(target_names, names)
    ddd <- which(unnamed == "...")
    if (length(ddd) > 0L) {
      stopifnot(length(ddd) == 1L)
      unnamed <- unnamed[seq_len(ddd - 1L)]
    }
    empty_idxs <- which(names == "")
    n <- min(length(empty_idxs), length(unnamed))
    if (n > 0L) names[empty_idxs[seq_len(n)]] <- unnamed[seq_len(n)]
    names <- c("", names)

    idx_x <- idx_l <- idx_f <- NULL

    if (fcn_name %in% c("xmap", "xmap_chr", "xmap_dbl", "xmap_int",
                          "xmap_lgl", "xmap_vec",
                          "xmap_dfc", "xmap_dfr",
                          "xmap_mat", "xmap_arr",
                          "xwalk",
                          "future_xmap", "future_xmap_chr",
                          "future_xmap_dbl", "future_xmap_int",
                          "future_xmap_lgl", "future_xmap_vec",
                          "future_xmap_dfc", "future_xmap_dfr",
                          "future_xmap_mat", "future_xmap_arr",
                          "future_xmap_raw",
                          "future_xwalk")) {
      ## xmap-style: .l + .f, cross-product iteration
      idx_l <- which(names == ".l")
      idx_f <- which(names == ".f")
    } else if (fcn_name %in% c("pmap_vec", "future_pmap_vec")) {
      ## pmap-style: .l + .f, parallel iteration
      idx_l <- which(names == ".l")
      idx_f <- which(names == ".f")
    } else if (fcn_name %in% c("map2_vec", "future_map2_vec")) {
      ## map2-style: .x + .f
      idx_x <- which(names == ".x")
      idx_f <- which(names == ".f")
    } else {
      ## map/imap-style: .x + .f
      idx_x <- which(names == ".x")
      idx_f <- which(names == ".f")
    }

    parts <- as.list(expr)

    if (!is.null(idx_x)) {
      stopifnot(length(idx_x) == 1L)
      parts[[idx_x]] <- bquote_apply(template_along, ALONG = parts[[idx_x]])
    }

    if (!is.null(idx_l)) {
      stopifnot(length(idx_l) == 1L)
      if (fcn_name %in% c("xmap", "xmap_chr", "xmap_dbl", "xmap_int",
                            "xmap_lgl", "xmap_vec",
                            "xmap_dfc", "xmap_dfr",
                            "xmap_mat", "xmap_arr",
                            "xwalk",
                            "future_xmap", "future_xmap_chr",
                            "future_xmap_dbl", "future_xmap_int",
                            "future_xmap_lgl", "future_xmap_vec",
                            "future_xmap_dfc", "future_xmap_dfr",
                            "future_xmap_mat", "future_xmap_arr",
                            "future_xmap_raw",
                            "future_xwalk")) {
        ## xmap: cross-product, so steps = prod(lengths(.l))
        parts[[idx_l]] <- bquote_apply(template_steps_prod_lengths,
                                        DATA = parts[[idx_l]])
      } else {
        ## pmap_vec: parallel, so along = .l[[1]]
        parts[[idx_l]] <- bquote_apply(template_along_first,
                                        ALONG = parts[[idx_l]])
      }
    }

    if (!is.null(idx_f)) {
      stopifnot(length(idx_f) == 1L)
      FUN <- expr[[idx_f]]
      parts[[idx_f]] <- bquote_apply(template_f, FUN = FUN)

      progressr_args <- list(
        .progressr_progressor = quote(.progressr_progressor)
      )
      parts <- c(parts, progressr_args)
    }

    bquote(local(.(as.call(parts))))
  } ## progressify_crossmap()
})


append_builtin_transpilers_for_crossmap <- local({
  known_fcns <- list(
    ## purrr-extensions
    imap_vec = c,
    map_vec = c,
    map2_vec = c,
    pmap_vec = c,
    ## xmap family
    xmap = c,
    xmap_chr = c,
    xmap_dbl = c,
    xmap_int = c,
    xmap_lgl = c,
    xmap_vec = c,
    xmap_dfc = c,
    xmap_dfr = c,
    xmap_mat = c,
    xmap_arr = c,
    xwalk = c,
    ## future purrr-extensions
    future_imap_vec = c,
    future_map_vec = c,
    future_map2_vec = c,
    future_pmap_vec = c,
    ## future xmap family
    future_xmap = c,
    future_xmap_chr = c,
    future_xmap_dbl = c,
    future_xmap_int = c,
    future_xmap_lgl = c,
    future_xmap_vec = c,
    future_xmap_dfc = c,
    future_xmap_dfr = c,
    future_xmap_mat = c,
    future_xmap_arr = c,
    future_xmap_raw = c,
    future_xwalk = c
  )

  template <- bquote_compile(function(expr, options) {
    ns <- getNamespace("crossmap")
    fcn <- get(.(fcn_name), mode = "function", envir = ns)
    progressify_crossmap(expr, fcn_name = .(fcn_name), fcn = fcn, envir = parent.frame())
  })

  make_transpiler <- function(fcn_name) {
    transpiler <- eval(bquote_apply(template))
    eval(transpiler)
  }

  function() {
    ## crossmap::xmap(), ...
    transpilers <- list()
    for (fcn_name in names(known_fcns)) {
      transpilers[[fcn_name]] <- list(
        label = sprintf("crossmap::%s() transpiler", fcn_name),
        transpiler = make_transpiler(fcn_name)
      )
    } ## for (fcn_name ...)
    transpilers <- list(crossmap = transpilers)

    append_transpilers("progressify::built-in", transpilers)

    ## Return required packages
    c("crossmap", "progressr")
  }
})

#' Evaluate a regular map-reduce call with progress updates
#'
#' @param expr An R expression.
#'
#' @param substitute If TRUE, `expr` is quoted.
#'
#' @param \ldots Not used.
#'
#' @param when If TRUE (default), the expression is progressified, otherwise not.
#'
#' @param eval If TRUE (default), the progressified expression is evaluated,
#' otherwise it is returned.
#'
#' @param envir The environment in which `expr` is evaluated.
#' 
#' @returns
#' Returns the value of the evaluated expression `expr`.
#'
#' @example incl/progressify-base.R
#'
#' @aliases pz
#' @importFrom progressr progressor
#' @export
progressify <- function(expr, substitute = TRUE, ..., when = TRUE, eval = TRUE, envir = parent.frame()) {
  if (substitute) expr <- substitute(expr)
  debug <- isTRUE(getOption("progressify.debug"))
  if (debug) {
    mdebug_push("progressify() ...")
    on.exit(mdebug_pop())
  }

  transpile(expr, substitute = FALSE, when = when, eval = eval, type = "progressify::built-in", envir = envir, what = "progressify", debug = debug)
} ## progressify()
class(progressify) <- c("transpiler", class(progressify))

#' @export
pz <- progressify

# lapply(X = xs, FUN = FUN, ...) =>
#
# local(
#   lapply(X = (.progressr_along <- xs), FUN = function(..., .progressr_progressor) {
#     on.exit(.progressr_progressor())
#     FUN(...)
#   }, .progressr_progressor = progressr::progressor(along = .progressr_along))
# )
#
progressify_base <- local({
  function(expr, fcn_name, fcn, ..., envir = parent.frame()) {
    names <- names(expr)
    if (is.null(names)) names <- rep("", length.out = length(expr))
    names <- names[-1]
    target_names <- names(formals(fcn))[seq_along(names)]
    unnamed <- setdiff(target_names, names)
    names[names == ""] <- unnamed
    names <- c("", names)

    idx_X <- idx_data <- idx_FUN <- idx_n <- idx_expr <- NULL

    if (fcn_name %in% c("by")) {
      idx_data <- which(names == "data")
      idx_FUN <- which(names == "FUN")
    } else if (fcn_name %in% c("eapply")) {
      idx_X <- which(names == "env")
      idx_FUN <- which(names == "FUN")
    } else if (fcn_name %in% c("lapply", "sapply", "vapply")) {
      idx_X <- which(names == "X")
      idx_FUN <- which(names == "FUN")
    } else if (fcn_name %in% c("replicate")) {
      idx_n <- which(names == "n")
      idx_expr <- which(names == "expr")
    } else {
      idx_FUN <- which(names == "FUN")
    }

    parts <- as.list(expr)

    if (!is.null(idx_X)) {
      stopifnot(length(idx_X) == 1L)
      parts[[idx_X]] <- bquote_apply(template_along, ALONG = parts[[idx_X]])
    }

    if (!is.null(idx_data)) {
      stopifnot(length(idx_data) == 1L)
      parts[[idx_data]] <- bquote_apply(template_steps_nrow, DATA = parts[[idx_data]])
    }

    if (!is.null(idx_FUN)) {
      stopifnot(length(idx_FUN) == 1L)
      FUN <- expr[[idx_FUN]]
      parts[[idx_FUN]] <- bquote_apply(template_FUN, FUN = FUN)

      progressr_args <- list(
        ...FUN = FUN,
        .progressr_progressor = quote(.progressr_progressor)
      )
      parts <- c(parts, progressr_args)
    }

    if (!is.null(idx_n)) {
      stopifnot(length(idx_n) == 1L)
      parts[[idx_n]] <- bquote_apply(template_steps, STEPS = parts[[idx_n]])
    }

    if (!is.null(idx_expr)) {
      stopifnot(length(idx_expr) == 1L)
      parts[[idx_expr]] <- bquote_apply(template_expr, EXPR = expr[[idx_expr]])
    }

    bquote(local(.(as.call(parts))))
  } ## progressify_base()
})


append_builtin_transpilers_for_base <- local({
  known_fcns <- list(
    apply = c,
    by = c,
    eapply = c,             ## done
    lapply = c,             ## done
    .mapply = c,
    mapply = c,
    Map = c,
    replicate = c,          ## done
    sapply = c,             ## done
    tapply = c,
    vapply = c              ## done
  )

  template <- bquote_compile(function(expr, options) {
    ns <- baseenv()
    fcn <- get(.(fcn_name), mode = "function", envir = ns)
    progressify_base(expr, fcn_name = .(fcn_name), fcn = fcn, envir = parent.frame())
  })

  make_transpiler <- function(fcn_name) {
    transpiler <- eval(bquote_apply(template))
    eval(transpiler)
  }

  function() {
    ## base::apply(), ...
    transpilers <- list()
    for (fcn_name in names(known_fcns)) {
      transpilers[[fcn_name]] <- list(
        label = sprintf("base::%s() transpiler", fcn_name),
        transpiler = make_transpiler(fcn_name)
      )
    } ## for (fcn_name ...)
    transpilers <- list(base = transpilers)

    append_transpilers("progressify::built-in", transpilers)

    ## Return required packages
    c("base", "progressr")
  }
})

#' Parse an R call to be transpiled
#'
#' @param call An R symbol or an R call.
#'
#' @param envir The environment from which to search for the function that
#' is called in the `call`.
#'
#' @param what A character string used in error messages describing what
#' type of transpiler is used.
#'
#' @param debug If TRUE, output debug information.
#'
#' @return
#' A named list with elements:
#'
#'  * `fcn`: the function called in the `call`
#'
#'  * `fcn_name`: the name of the function
#'
#'  * `ns_name`: the name of the namespace where the function lives
#'
#' @keywords internal
#' @noRd
parse_call <- function(call, envir = parent.frame(), what = "transpiler", debug = FALSE) {
  if (debug) {
    mdebug_push("parse_call() ...")
    on.exit(mdebug_pop())
  }
  
  if (is.symbol(call)) {
    ## Examples: lapply(...), map(...)
    namespace <- NULL
    fcn <- call
  } else if (is.call(call)) {
    stopifnot(length(call) == 3L)
    ## Examples: base::lapply(...), purrr::map(...)
    ## Not supported: base:::lapply(), baseenv()$lapply(...), ...
    op <- call[[1]]
    if (is.symbol(op)) {
      op_name <- as.character(op)
      if (op_name == "::") {
      } else if (op_name == ":::") {
        stop(sprintf("Do not know how to %s a private function: %s()", what, deparse(call)))
      } else {
        stop(sprintf("Do not know how to %s function %s()", what, deparse(call)))
      }
      namespace <- call[[2]]
      fcn <- call[[3]]
    } else {
      stop(sprintf("Do not know how to %s object of type %s: %s()", what, sQuote(typeof(op)), deparse(call)))
    }
  } else {
    stop(sprintf("Do not know how to %s expression: %s", what, as.character(call)))
  }

  ns_name <- as.character(namespace)
  fcn_name <- as.character(fcn)
  if (debug) {
    if (length(ns_name) == 1L) {
      msg <- sprintf("Function: %s::%s(...)", ns_name, fcn_name)
    } else {
      msg <- sprintf("Function: %s(...)", fcn_name)
    }
    mdebug(msg)
  }

  ## Does the function exist?
  if (is.null(namespace)) {
    if (debug) mdebug_push("Locate function ...")
    if (!exists(fcn_name, mode = "function", envir = envir, inherits = TRUE)) {
      name <- deparse(call)
      if (grepl("^%.*%$", name)) {
        stop(sprintf("Unknown infix operator: %s", name))
      } else {
        stop(sprintf("Unknown function: %s()", name))
      }
    }
    fcn <- get(fcn_name, mode = "function", envir = envir, inherits = TRUE)
    env <- environment(fcn)

    if (is.null(env) && is.primitive(fcn)) {
      env <- baseenv()
    } else if (inherits(fcn, "standardGeneric")) {
      env <- parent.env(env)
    }
    tenv <- env
    repeat {
      ns_name <- environmentName(tenv)
      if (nzchar(ns_name)) break
      tenv_next <- parent.env(tenv)
      if (identical(tenv_next, tenv)) {
        stop(sprintf("Internal error: infinite environment stack in parse_call(): %s", sQuote(ns_name)))
      }
      tenv <- tenv_next
    }
    
    stopifnot(nzchar(ns_name))
    if (debug) {
      mdebugf("Function located in: %s", sQuote(ns_name))
      mdebug_pop()
    }
  } else {
    ns <- getNamespace(ns_name)
    if (!exists(fcn_name, mode = "function", envir = ns, inherits = TRUE)) {
      stop(sprintf("No such function in %s: %s()", sQuote(ns_name), deparse(call)))
    }
    fcn <- get(fcn_name, mode = "function", envir = ns, inherits = TRUE)
  }

  list(fcn = fcn, ns_name = ns_name, fcn_name = fcn_name)
} ## parse_call()




#' Append arguments to a call.
#'
#' @param expr An \R call expression.
#'
#' @param \ldots,.args Named arguments to be appended to the call expression.
#'
#' @return
#' Return the expression with arguments appended.
#'
#' @examples
#' call <- quote(my_fcn(x, y))
#' print(call)
#' #> my_fcn(x, y)
#'
#' call2 <- append_call_arguments(call, z = 42, w = quote(1 + 2))
#' print(call2)
#' #> my_fcn(x, y, z = 42, w = 1 + 2)
#'
#' @noRd
append_call_arguments <- function(expr, ..., .args = list(...)) {
  ## Update 'EXPR'
  as.call(c(
    as.list(expr),
    .args
  ))
} ## append_call_arguments()

descend_wrappers <- function(expr, envir = parent.frame(), unwrap, what = "unwrap", debug = FALSE) {
  ## Nothing to do?
  if (length(unwrap) == 0 || !is.call(expr)) return(1L)

  if (debug) {
    mdebug_push("descend_wrappers() ...")
    on.exit(mdebug_pop())
  }

  call <- expr[[1]]  ## e.g. {, local, lapply, ...
  if (debug) {
    mdebug("Call:")
    mprint(call)
  }

  call_info <- parse_call(call, envir = envir, what = what, debug = debug)
  fcn <- call_info[["fcn"]]
  fcn_name <- call_info[["fcn_name"]]
  ns_name <- call_info[["ns_name"]]

  for (wrapper in unwrap) {
    if (identical(fcn, wrapper)) {  ## a wrapper call?
      if (debug) {
        info <- switch(fcn_name,
          "{" = "{ ... }",
          "(" = "( ... )",
          sprintf("%s( ... )", fcn_name)
        )
        mdebugf("Wrapped in %s", info)
      }

      ## Which argument is the wrapped expression?
      index <- if (identical(fcn, base::`{`)) {
        length(expr)
      } else if (identical(fcn, base::with)) {
        ## with(data, expr) - the expression is the second argument (position 3)
        3L
      } else {
        2L
      }

      ## Safety check
      if (index > length(expr)) return(1L)

      return(c(index, descend_wrappers(expr[[index]], envir = envir, unwrap = unwrap, what = what, debug = debug)))
    } ## if (identical(fcn, wrapper)
  } ## for (wrapper in unwrap)

  return(1L)
} ## descend_wrappers()



#' @export
print.transpiled_call <- function(x, ...) {
  stopifnot(inherits(x, "call"))

  ...x_org... <- x

  ## Make sure attributes are displayed
  ## Whoa; this might be more convoluted than writing a custom
  ## print() method from scratch
  ...x... <- as.list(x)
  for (...kk... in seq_along(...x...)) {
    ...x...[[...kk...]] <- local({
      expr <- ...x...[[...kk...]]
      code <- c("substitute(", deparse(expr), ")")
      tryCatch({
        eval(parse(text = code), enclos = emptyenv())
      }, error = function(e) {
        ## Some deparsed expression cannot be parsed
        expr
      })
    })
  }
  x <- as.call(...x...)
  
  NextMethod()

  invisible(...x_org...)
}

#' Finds registered S3 method for S3 generic function and object
#'
#' @param fcn,fcn_name An S3 generic function and its name.
#'
#' @param call The S3 function call, which includes the dispatch object.
#'
#' @param envir The environment in which the dispatch object should be
#' resolved.
#'
#' @param debug If TRUE, debug output is given.
#'
#' @return
#' Returns a named list of elements `package` and `name` if found,
#' otherwise NULL.
#'
#' @noRd
#' @importFrom utils getS3method
find_s3_method <- function(fcn, fcn_name, call, envir, debug = FALSE) {
  ## Get the name of the first argument, which is the S3 dispatch argument
  fmls <- formals(fcn)
  if (length(fmls) == 0L) return(NULL)

  ## FIXME: Here we assume we're dispatching on the first argument
  dispatch_arg_name <- names(fmls)[[1L]]
  if (dispatch_arg_name == "...") return(NULL) ## FIXME: Skip for now

  ## Use match.call() to correctly handle named and reordered arguments
  matched_call <- tryCatch(
    match.call(fcn, call = call),
    error = function(e) NULL
  )
  if (is.null(matched_call)) return(NULL)

  dispatch_expr <- matched_call[[dispatch_arg_name]]
  if (is.null(dispatch_expr)) return(NULL)
  if (!is.symbol(dispatch_expr) && !is.call(dispatch_expr)) return(NULL)

  ## Evaluate the dispatch argument to get its class
  dispatch_obj <- tryCatch(
    eval(dispatch_expr, envir = envir),
    error = function(e) NULL
  )
  if (is.null(dispatch_obj)) return(NULL)

  ## Use .class2() to get the full S3 dispatch chain, which includes inherited
  ## classes from S4/R5 hierarchies not visible in class() alone.
  ## Example: class(lmerMod_obj) = "lmerMod", but .class2() = c("lmerMod", "merMod")
  dispatch_classes <- .class2(dispatch_obj)

  ## Walk the class hierarchy to find a dispatched S3 method
  method <- NULL
  dispatch_class <- NULL
  for (cls in dispatch_classes) {
    m <- getS3method(fcn_name, cls, optional = TRUE)
    if (!is.null(m)) {
      method <- m
      dispatch_class <- cls
      break
    }
  }
  if (is.null(method)) return(NULL)

  ## Determine the package the method lives in
  method_env <- environment(method)
  if (is.null(method_env)) return(NULL)
  method_pkg <- environmentName(topenv(method_env))

  method_name <- paste0(fcn_name, ".", dispatch_class)

  if (debug) {
    mdebugf("S3 generic %s() dispatches to %s::%s() for class %s",
            fcn_name, method_pkg, method_name, sQuote(dispatch_class))
  }

  list(package = method_pkg, name = method_name)
} ## find_s3_method()

## This function registered functions that adds transpilers for specific
## packages, without loading those packages.
register_all_transpilers <- function() {
  ## Map-reduce packages (base-R)
  append_builtin_transpilers_for_base()
  append_builtin_transpilers_for_boot()
  append_builtin_transpilers_for_crossmap()
  append_builtin_transpilers_for_future.apply()
  append_builtin_transpilers_for_fwb()
  append_builtin_transpilers_for_purrr()
  append_builtin_transpilers_for_furrr()
  append_builtin_transpilers_for_partykit()
  append_builtin_transpilers_for_plyr()
  append_builtin_transpilers_for_sandwich()
  append_builtin_transpilers_for_stats()
  append_builtin_transpilers_for_foreach()
  append_builtin_transpilers_for_doFuture()
  append_builtin_transpilers_for_lme4()
  append_builtin_transpilers_for_SimDesign()
} ## register_all_transpilers()

#' Finds S4 method package for S4 generic function and object
#'
#' @param fcn,fcn_name An S4 generic function and its name.
#'
#' @param call The S4 function call, which includes the dispatch object.
#'
#' @param envir The environment in which the dispatch object should be
#' resolved.
#'
#' @param debug If TRUE, debug output is given.
#'
#' @return
#' Returns a named list of elements `package` and `name` if found,
#' otherwise NULL.
#'
#' @noRd
find_s4_method <- function(fcn, fcn_name, call, envir, debug = FALSE) {
  if (!methods::isGeneric(fcn_name)) return(NULL)

  ## Get the name of the first argument (dispatch argument)
  fmls <- formals(fcn)
  if (length(fmls) == 0L) return(NULL)

  ## FIXME: Here we assume we're dispatching on the first argument
  dispatch_arg_name <- names(fmls)[[1L]]
  if (dispatch_arg_name == "...") return(NULL) ## FIXME: Skip for now

  ## Use match.call() to correctly handle named and reordered arguments
  matched_call <- tryCatch(
    match.call(fcn, call = call),
    error = function(e) NULL
  )
  if (is.null(matched_call)) return(NULL)

  dispatch_expr <- matched_call[[dispatch_arg_name]]
  if (is.null(dispatch_expr)) return(NULL)
  if (!is.symbol(dispatch_expr) && !is.call(dispatch_expr)) return(NULL)

  ## Evaluate the dispatch argument to get its class
  dispatch_obj <- tryCatch(
    eval(dispatch_expr, envir = envir),
    error = function(e) NULL
  )
  if (is.null(dispatch_obj)) return(NULL)

  dispatch_class <- class(dispatch_obj)[[1L]]

  ## Find the S4 method
  method <- tryCatch(
    methods::selectMethod(fcn_name, signature = dispatch_class),
    error = function(e) NULL
  )
  if (is.null(method)) return(NULL)

  ## Determine the package the method is defined in
  method_env <- environment(method)
  if (is.null(method_env)) return(NULL)
  method_pkg <- environmentName(topenv(method_env))

  if (debug) {
    mdebugf("S4 generic %s() dispatches to %s::%s() for class %s",
            fcn_name, method_pkg, fcn_name, sQuote(dispatch_class))
  }

  list(package = method_pkg, name = fcn_name)
} ## find_s4_method()

# SimDesign::runSimulation(...) =>
#
# local({
#   design <- MyDesign
#   replications <- 10
#   .progressr_steps <- if (length(replications) == 1L) replications * nrow(design) else sum(replications)
#   .progressr_progressor <- progressr::progressor(steps = .progressr_steps)
#   .progressr_analyse <- Anal
#   SimDesign::runSimulation(
#     design = design,
#     replications = replications,
#     generate = generate,
#     analyse = function(condition, dat, fixed_objects = NULL, ...) {
#       on.exit(.progressr_progressor())
#       .progressr_analyse(condition = condition, dat = dat, fixed_objects = fixed_objects, ...)
#     },
#     summarise = summarise,
#     progress = FALSE
#   )
# })
#
progressify_SimDesign <- local({
  template_runSimulation_outer <- bquote_compile(local({
    design <- .(DESIGN)
    replications <- .(REPLICATIONS)
    .progressr_steps <- if (length(replications) == 1L) replications * nrow(design) else sum(replications)
    .progressr_progressor <- progressr::progressor(steps = .progressr_steps)
    .progressr_analyse <- .(ANALYSE)
    .(EXPR)
  }))

  template_runSimulation_FUN <- quote(function(condition, dat, fixed_objects = NULL, ...) {
    on.exit(.progressr_progressor())
    .progressr_analyse(condition = condition, dat = dat, fixed_objects = fixed_objects, ...)
  })

  function(expr, fcn_name, fcn, ..., envir = parent.frame()) {
    names <- names(expr)
    if (is.null(names)) names <- rep("", length.out = length(expr))
    names <- names[-1]
    target_names <- names(formals(fcn))[seq_along(names)]
    unnamed <- setdiff(target_names, names)
    ## Handle '...' in formals - only resolve positional args before it
    ddd <- which(unnamed == "...")
    if (length(ddd) > 0L) {
      stopifnot(length(ddd) == 1L)
      unnamed <- unnamed[seq_len(ddd - 1L)]
    }
    empty_idxs <- which(names == "")
    n <- min(length(empty_idxs), length(unnamed))
    if (n > 0L) names[empty_idxs[seq_len(n)]] <- unnamed[seq_len(n)]
    names <- c("", names)

    idx_design <- which(names == "design")
    idx_replications <- which(names == "replications")
    idx_analyse <- which(names == "analyse")

    parts <- as.list(expr)

    stopifnot(length(idx_design) == 1L)
    design <- parts[[idx_design]]
    parts[[idx_design]] <- quote(design)

    stopifnot(length(idx_replications) == 1L)
    replications <- parts[[idx_replications]]
    parts[[idx_replications]] <- quote(replications)

    stopifnot(length(idx_analyse) == 1L)
    analyse <- parts[[idx_analyse]]
    parts[[idx_analyse]] <- template_runSimulation_FUN

    # Inject progress = FALSE to disable SimDesign's native console progress bar
    idx_progress <- which(names == "progress")
    if (length(idx_progress) == 0L) {
      parts$progress <- FALSE
    }

    bquote_apply(template_runSimulation_outer,
                 DESIGN = design,
                 REPLICATIONS = replications,
                 ANALYSE = analyse,
                 EXPR = as.call(parts))
  } ## progressify_SimDesign()
})


append_builtin_transpilers_for_SimDesign <- local({
  known_fcns <- list(
    runSimulation = c
  )

  template <- bquote_compile(function(expr, options) {
    ns <- getNamespace("SimDesign")
    fcn <- get(.(fcn_name), mode = "function", envir = ns)
    progressify_SimDesign(expr, fcn_name = .(fcn_name), fcn = fcn, envir = parent.frame())
  })

  make_transpiler <- function(fcn_name) {
    transpiler <- eval(bquote_apply(template))
    eval(transpiler)
  }

  function() {
    ## SimDesign::runSimulation()
    transpilers <- list()
    for (fcn_name in names(known_fcns)) {
      transpilers[[fcn_name]] <- list(
        label = sprintf("SimDesign::%s() transpiler", fcn_name),
        transpiler = make_transpiler(fcn_name)
      )
    } ## for (fcn_name ...)
    transpilers <- list(SimDesign = transpilers)

    append_transpilers("progressify::built-in", transpilers)

    ## Return required packages
    c("SimDesign", "progressr")
  }
})

<!--
%\VignetteIndexEntry{Progress updates for 'furrr' functions}
%\VignetteAuthor{Henrik Bengtsson}
%\VignetteKeyword{R}
%\VignetteKeyword{package}
%\VignetteKeyword{furrr}
%\VignetteKeyword{vignette}
%\VignetteKeyword{progressify}
%\VignetteEngine{progressify::selfonly}
-->

The **progressify** package allows you to easily add progress
reporting to sequential and parallel map-reduce code by piping to the
`progressify()` function. Easy!


# TL;DR

```r
library(progressify)
handlers(global = TRUE)
library(furrr)
plan(multisession)

slow_fcn <- function(x) {
  Sys.sleep(0.1)  # emulate work
  x^2
}

xs <- 1:100
ys <- xs |> future_map(slow_fcn) |> progressify()
```


# Introduction

This vignette demonstrates how to use this approach to add progress
reporting to **[furrr]** functions such as `future_map()`,
`future_map_dbl()`, and `future_walk()`.

The **furrr** `future_map()` function is commonly used to apply a
function to the elements of a vector or a list in parallel. For
example,

```r
library(furrr)
plan(multisession)

xs <- 1:100
ys <- xs |> future_map(slow_fcn)
```

Here `future_map()` provides no feedback on how far it has progressed,
but we can easily add progress reporting by using:

```r
library(furrr)
plan(multisession)

library(progressify)
handlers(global = TRUE)

xs <- 1:100
ys <- xs |> future_map(slow_fcn) |> progressify()
```

Using the default progress handler, the progress reporting will
appear as:

```plain
  |=====                    |  20%
```


# Supported Functions

The `progressify()` function supports the following **furrr** functions:

 * `future_map()`, `future_map_chr()`, `future_map_dbl()`, `future_map_int()`, `future_map_lgl()`, `future_map_raw()`, `future_map_vec()`, `future_map_dfr()`, `future_map_dfc()`, `future_map_at()`, `future_map_if()`, `future_walk()`
 * `future_map2()`, `future_map2_chr()`, `future_map2_dbl()`, `future_map2_int()`, `future_map2_lgl()`, `future_map2_raw()`, `future_map2_vec()`, `future_map2_dfr()`, `future_map2_dfc()`, `future_walk2()`
 * `future_pmap()`, `future_pmap_chr()`, `future_pmap_dbl()`, `future_pmap_int()`, `future_pmap_lgl()`, `future_pmap_raw()`, `future_pmap_vec()`, `future_pmap_dfr()`, `future_pmap_dfc()`, `future_pwalk()`
 * `future_imap()`, `future_imap_chr()`, `future_imap_dbl()`, `future_imap_int()`, `future_imap_lgl()`, `future_imap_raw()`, `future_imap_vec()`, `future_imap_dfr()`, `future_imap_dfc()`, `future_iwalk()`
 * `future_modify()`, `future_modify_at()`, `future_modify_if()`


[furrr]: https://cran.r-project.org/package=furrr

<!--
%\VignetteIndexEntry{Progress updates for 'fwb' functions}
%\VignetteAuthor{Henrik Bengtsson}
%\VignetteKeyword{R}
%\VignetteKeyword{package}
%\VignetteKeyword{fwb}
%\VignetteKeyword{vignette}
%\VignetteKeyword{progressify}
%\VignetteEngine{progressify::selfonly}
-->

The **progressify** package allows you to easily add progress
reporting to sequential and parallel map-reduce code by piping to the
`progressify()` function. Easy!


# TL;DR

```r
library(progressify)
handlers(global = TRUE)
library(fwb)

# Run fractional weighted bootstrap with progress signaling
my_stat <- function(data, w) coef(lm(mpg ~ cyl, data = data, weights = w))
res <- fwb(data = mtcars, statistic = my_stat, R = 1000) |> progressify()
```


# Introduction

This vignette demonstrates how to use this approach to add progress
reporting to the **[fwb]** package's main function `fwb()`.

The **fwb** package provides functions for generating fractional weighted bootstrap replicates.
For example, `fwb()` runs a statistic function `R` times:

```r
library(fwb)
my_stat <- function(data, w) coef(lm(mpg ~ cyl, data = data, weights = w))
res <- fwb(data = mtcars, statistic = my_stat, R = 1000)
```

By default, `fwb()` uses `verbose = TRUE`, which provides progress
feedback via the **[pbapply]** package, where the style can be
controlled via `pbapply::pboptions()`.

As an alternative, we can use the `progressify()` function to report
on progress via any combination of **progressr** reporters. To do
this, use:

```r
library(fwb)

library(progressify)
handlers(global = TRUE)

my_stat <- function(data, w) coef(lm(mpg ~ cyl, data = data, weights = w))
res <- fwb(data = mtcars, statistic = my_stat, R = 1000) |> progressify()
```

Comment: This will disable the built-in progress feedback by setting
`verbose = FALSE` in order to avoid dual reporting.


# Supported Functions

The `progressify()` function supports the following **fwb**
functions:

 * `fwb()`


[fwb]: https://cran.r-project.org/package=fwb
[pbapply]: https://cran.r-project.org/package=pbapply

<!--
%\VignetteIndexEntry{Progress updates for 'purrr' functions}
%\VignetteAuthor{Henrik Bengtsson}
%\VignetteKeyword{R}
%\VignetteKeyword{package}
%\VignetteKeyword{purrr}
%\VignetteKeyword{vignette}
%\VignetteKeyword{progressify}
%\VignetteEngine{progressify::selfonly}
-->

The **progressify** package allows you to easily add progress
reporting to sequential and parallel map-reduce code by piping to the
`progressify()` function. Easy!


# TL;DR

```r
library(progressify)
handlers(global = TRUE)
library(purrr)

slow_fcn <- function(x) {
  Sys.sleep(0.1)  # emulate work
  x^2
}

xs <- 1:100
ys <- xs |> map(slow_fcn) |> progressify()
```


# Introduction

This vignette demonstrates how to use this approach to add progress
reporting to **[purrr]** functions such as `map()`, `map_dbl()`, and
`walk()`.

The **purrr** `map()` function is commonly used to apply a function to
the elements of a vector or a list. For example,

```r
library(purrr)
xs <- 1:100
ys <- map(xs, slow_fcn)
```

or equivalently using pipe syntax

```r
xs <- 1:100
ys <- xs |> map(slow_fcn)
```

Here `map()` provides no feedback on how far it has progressed,
but we can easily add progress reporting by using:

```r
library(purrr)

library(progressify)
handlers(global = TRUE)

xs <- 1:100
ys <- xs |> map(slow_fcn) |> progressify()
```

Using the default progress handler, the progress reporting will
appear as:

```plain
  |=====                    |  20%
```


# Supported Functions

The `progressify()` function supports the following **purrr** functions:

 * `map()`, `map_chr()`, `map_dbl()`, `map_int()`, `map_lgl()`, `map_vec()`, `walk()`
 * `map2()`, `map2_chr()`, `map2_dbl()`, `map2_int()`, `map2_lgl()`, `map2_vec()`, `walk2()`
 * `pmap()`, `pmap_chr()`, `pmap_dbl()`, `pmap_int()`, `pmap_lgl()`, `pmap_vec()`, `pwalk()`
 * `imap()`, `imap_chr()`, `imap_dbl()`, `imap_int()`, `imap_lgl()`, `imap_vec()`
 * `modify()`, `modify2()`, `imodify()`


[purrr]: https://cran.r-project.org/package=purrr

<!--
%\VignetteIndexEntry{Progress updates for 'lme4' functions}
%\VignetteAuthor{Henrik Bengtsson}
%\VignetteKeyword{R}
%\VignetteKeyword{package}
%\VignetteKeyword{lme4}
%\VignetteKeyword{vignette}
%\VignetteKeyword{progressify}
%\VignetteEngine{progressify::selfonly}
-->

The **progressify** package allows you to easily add progress
reporting to sequential and parallel map-reduce code by piping to the
`progressify()` function. Easy!


# TL;DR

```r
library(progressify)
handlers(global = TRUE)
library(lme4)

# Fit random-slope model
fm1 <- lmer(Reaction ~ Days + (Days | Subject), sleepstudy)
my_stat <- function(fit) {
  fixef(fit)
}

# Run bootstrap with progress signaling
res <- bootMer(fm1, my_stat, nsim = 1000) |> progressify()
```


# Introduction

This vignette demonstrates how to use this approach to add progress
reporting to **[lme4]** functions such as `bootMer()`. The **lme4**
package provides functions for fitting linear, generalized linear, and
nonlinear mixed-effects models. For example, `bootMer()` runs a
statistic function `nsim` times:

```r
library(lme4)

# Fit random-slope model
fm1 <- lmer(Reaction ~ Days + (Days | Subject), sleepstudy)
my_stat <- function(fit) {
  fixef(fit)
}

res <- bootMer(fm1, my_stat, nsim = 1000)
```

By default, `bootMer()` provides no progress feedback. However, we can
easily add progress reporting using the `progressify()` function:

```r
library(lme4)
library(progressify)
handlers(global = TRUE)

# Fit random-slope model
fm1 <- lmer(Reaction ~ Days + (Days | Subject), sleepstudy)
my_stat <- function(fit) {
  fixef(fit)
}

res <- bootMer(fm1, my_stat, nsim = 1000) |> progressify()
```


# Supported Functions

The `progressify()` function supports the following **lme4**
functions:

 * `bootMer()`


[lme4]: https://cran.r-project.org/package=lme4

<!--
%\VignetteIndexEntry{Progress updates for 'stats' functions}
%\VignetteAuthor{Henrik Bengtsson}
%\VignetteKeyword{R}
%\VignetteKeyword{package}
%\VignetteKeyword{stats}
%\VignetteKeyword{vignette}
%\VignetteKeyword{progressify}
%\VignetteEngine{progressify::selfonly}
-->

The **progressify** package allows you to easily add progress
reporting to sequential and parallel map-reduce code by piping to the
`progressify()` function. Easy!


# TL;DR

```r
library(progressify)
handlers(global = TRUE)
library(stats)

d <- as.dendrogram(hclust(dist(USArrests)))
d2 <- dendrapply(d, function(n) { Sys.sleep(0.01); n }) |> progressify()
```


# Introduction

This vignette demonstrates how to use this approach to add progress
reporting to functions such as `dendrapply()` in the **stats**
package. For example, consider the `dendrapply()` function, which is
commonly used to apply a function to the nodes of a dendrogram, as in:

```r
d <- as.dendrogram(hclust(dist(USArrests)))
d2 <- dendrapply(d, function(n) { Sys.sleep(0.01); n })
```

Here `dendrapply()` provides no feedback on how far it has
progressed, but we can easily add progress reporting by using:

```r
library(progressify)
handlers(global = TRUE)

d2 <- dendrapply(d, function(n) { Sys.sleep(0.01); n }) |> progressify()
```

Using the default progress handler, the progress reporting will
appear as:

```plain
  |=====                    |  20%
```


# Supported Functions

The `progressify()` function supports the following **stats** package
functions:

 * `dendrapply()`

<!--
%\VignetteIndexEntry{Progress updates for 'plyr' functions}
%\VignetteAuthor{Henrik Bengtsson}
%\VignetteKeyword{R}
%\VignetteKeyword{package}
%\VignetteKeyword{plyr}
%\VignetteKeyword{vignette}
%\VignetteKeyword{progressify}
%\VignetteEngine{progressify::selfonly}
-->

The **progressify** package allows you to easily add progress
reporting to sequential and parallel map-reduce code by piping to the
`progressify()` function. Easy!


# TL;DR

```r
library(progressify)
handlers(global = TRUE)
library(plyr)

slow_fcn <- function(x) {
  Sys.sleep(0.1)  # emulate work
  x^2
}

xs <- 1:100
ys <- llply(xs, slow_fcn) |> progressify()
```


# Introduction

This vignette demonstrates how to use this approach to add progress
reporting to **[plyr]** functions such as `llply()`, `maply()`, and
`ddply()`.

The **plyr** `llply()` function is commonly used to apply a function
to the elements of a list and return a list. For example,

```r
library(plyr)
xs <- 1:100
ys <- llply(xs, slow_fcn)
```

Here `llply()` provides no feedback on how far it has progressed,
but we can easily add progress reporting by using:

```r
library(plyr)

library(progressify)
handlers(global = TRUE)

xs <- 1:100
ys <- llply(xs, slow_fcn) |> progressify()
```

Using the default progress handler, the progress reporting will
appear as:

```plain
  |=====                    |  20%
```


# Supported Functions

The `progressify()` function supports the following **plyr**
functions:

 * `l_ply()`, `laply()`, `ldply()`, `llply()`
 * `m_ply()`, `maply()`, `mdply()`, `mlply()`
 * `r_ply()`, `raply()`, `rdply()`, `rlply()`


[plyr]: https://cran.r-project.org/package=plyr

<!--
%\VignetteIndexEntry{Progress updates for 'future.apply' functions}
%\VignetteAuthor{Henrik Bengtsson}
%\VignetteKeyword{R}
%\VignetteKeyword{package}
%\VignetteKeyword{future.apply}
%\VignetteKeyword{vignette}
%\VignetteKeyword{progressify}
%\VignetteEngine{progressify::selfonly}
-->

The **progressify** package allows you to easily add progress
reporting to sequential and parallel map-reduce code by piping to the
`progressify()` function. Easy!


# TL;DR

```r
library(progressify)
handlers(global = TRUE)
library(future.apply)
plan(multisession)

slow_fcn <- function(x) {
  Sys.sleep(0.1)  # emulate work
  x^2
}

xs <- 1:100
ys <- future_lapply(xs, slow_fcn) |> progressify()
```


# Introduction

This vignette demonstrates how to use this approach to add progress
reporting to **[future.apply]** functions such as `future_lapply()`,
`future_tapply()`, `future_apply()`, and `future_replicate()`.

The **future.apply** `future_lapply()` function is commonly used to
apply a function to the elements of a vector or a list in parallel.
For example,

```r
library(future.apply)
plan(multisession)

xs <- 1:100
ys <- future_lapply(xs, slow_fcn)
```

Here `future_lapply()` provides no feedback on how far it has
progressed, but we can easily add progress reporting by using:

```r
library(future.apply)
plan(multisession)

library(progressify)
handlers(global = TRUE)

ys <- future_lapply(xs, slow_fcn) |> progressify()
```

Using the default progress handler, the progress reporting will
appear as:

```plain
  |=====                    |  20%
```


# Supported Functions

The `progressify()` function supports the following **future.apply**
functions:

 * `future_lapply()`, `future_vapply()`, `future_sapply()`, `future_tapply()`
 * `future_mapply()`, `future_.mapply()`, `future_Map()`
 * `future_eapply()`
 * `future_apply()`
 * `future_replicate()`
 * `future_by()`


[future.apply]: https://cran.r-project.org/package=future.apply

<!--
%\VignetteIndexEntry{Progress updates for 'SimDesign' functions}
%\VignetteAuthor{Henrik Bengtsson}
%\VignetteKeyword{R}
%\VignetteKeyword{package}
%\VignetteKeyword{SimDesign}
%\VignetteKeyword{vignette}
%\VignetteKeyword{progressify}
%\VignetteEngine{progressify::selfonly}
-->

The **progressify** package allows you to easily add progress
reporting to sequential and parallel map-reduce code by piping to the
`progressify()` function. Easy!


# TL;DR

```r
library(progressify)
handlers(global = TRUE)
library(SimDesign)

# Create small design
Design <- createDesign(factor1 = c(1, 2))

Generate <- function(condition, fixed_objects = NULL) {
  rnorm(100)
}

Analyse <- function(condition, dat, fixed_objects = NULL) {
  mean(dat)
}

Summarise <- function(condition, results, fixed_objects = NULL) {
  mean(results)
}

# Run simulation with progress signaling
res <- runSimulation(design = Design, replications = 100,
                     generate = Generate, analyse = Analyse,
                     summarise = Summarise) |> progressify()
```


# Introduction

This vignette demonstrates how to use this approach to add progress
reporting to **[SimDesign]** functions such as `runSimulation()`.  The
**SimDesign** package provides a comprehensive framework for Monte
Carlo simulation studies in R. For example, `runSimulation()`
evaluates the `generate` and `analyse` functions `replications` times
for each condition in the `design`:

```r
library(SimDesign)

Design <- createDesign(factor1 = c(1, 2))
Generate <- function(condition, fixed_objects = NULL) rnorm(100)
Analyse <- function(condition, dat, fixed_objects = NULL) mean(dat)
Summarise <- function(condition, results, fixed_objects = NULL) mean(results)

res <- runSimulation(design = Design, replications = 100,
                     generate = Generate, analyse = Analyse,
                     summarise = Summarise)
```

By default, `runSimulation()` provides its own text/console-based
progress bar. However, we can easily replace this with
**progressr**-based reporting using the `progressify()` function:

```r
library(SimDesign)
library(progressify)
handlers(global = TRUE)

res <- runSimulation(design = Design, replications = 100,
                     generate = Generate, analyse = Analyse,
                     summarise = Summarise) |> progressify()
```

When progressified, `progressify()` automatically silences
`SimDesign`'s native console progress bar and routes the progress
updates through `progressr`. This means you can use any of the
[supported progressr handlers], e.g., Shiny progress bars, system
notifications, or CLI progress spinners.

# Supported Functions

The `progressify()` function supports the following **SimDesign** functions:

 * `runSimulation()`


[SimDesign]: https://cran.r-project.org/package=SimDesign
[supported progressr handlers]: https://progressr.futureverse.org/articles/progressr-11-handlers.html

<!--
%\VignetteIndexEntry{Progress updates for 'sandwich' functions}
%\VignetteAuthor{Henrik Bengtsson}
%\VignetteKeyword{R}
%\VignetteKeyword{package}
%\VignetteKeyword{sandwich}
%\VignetteKeyword{vignette}
%\VignetteKeyword{progressify}
%\VignetteEngine{progressify::selfonly}
-->

The **progressify** package allows you to easily add progress
reporting to sequential and parallel map-reduce code by piping to the
`progressify()` function. Easy!


# TL;DR

```r
library(progressify)
handlers(global = TRUE)
library(sandwich)

fit <- lm(dist ~ speed, data = cars)
v <- vcovBS(fit, R = 100L) |> progressify()
```


# Introduction

This vignette demonstrates how to use this approach to add progress
reporting to **[sandwich]** functions such as `vcovBS()` and `vcovJK()`.

The **sandwich** package provides model-robust standard error
estimators for cross-section, time series, and longitudinal data.
Some of these estimators, specifically the bootstrap and jackknife
estimators, are computationally intensive and can benefit from
progress reporting.

For example, `vcovBS()` computes bootstrapped covariance matrix
estimators.

```r
library(sandwich)
fit <- lm(dist ~ speed, data = cars)
v <- vcovBS(fit, R = 100L)
```

Here `vcovBS()` provides no feedback on how far it has progressed,
but we can easily add progress reporting by using:

```r
library(sandwich)

library(progressify)
handlers(global = TRUE)

fit <- lm(dist ~ speed, data = cars)
v <- vcovBS(fit, R = 100L) |> progressify()
```

Similarly, the jackknife estimator `vcovJK()` can be progressified:

```r
library(sandwich)

library(progressify)
handlers(global = TRUE)

fit <- lm(dist ~ speed, data = cars)
v <- vcovJK(fit) |> progressify()
```


# Supported Functions

The `progressify()` function supports the following **sandwich**
functions:

 * `vcovBS()`
 * `vcovJK()`


[sandwich]: https://cran.r-project.org/package=sandwich

<!--
%\VignetteIndexEntry{Progress updates for base-R apply functions}
%\VignetteAuthor{Henrik Bengtsson}
%\VignetteKeyword{R}
%\VignetteKeyword{package}
%\VignetteKeyword{vignette}
%\VignetteKeyword{progressify}
%\VignetteEngine{progressify::selfonly}
-->

The **progressify** package allows you to easily add progress
reporting to sequential and parallel map-reduce code by piping to the
`progressify()` function. Easy!


# TL;DR

```r
library(progressify)
handlers(global = TRUE)

slow_fcn <- function(x) {
  Sys.sleep(0.1)  # emulate work
  x^2
}

xs <- 1:100
ys <- lapply(xs, slow_fcn) |> progressify()
```


# Introduction

This vignette demonstrates how to use this approach to add progress
reporting to functions such as `lapply()`, `tapply()`, `apply()`, and
`replicate()` in the **base** package. For example, consider the base
R `lapply()` function, which is commonly used to apply a function to
the elements of a vector or a list, as in:

```r
xs <- 1:100
ys <- lapply(xs, slow_fcn)
```

Here `lapply()` provides no feedback on how far it has progressed,
but we can easily add progress reporting by using:

```r
library(progressify)
handlers(global = TRUE)

ys <- lapply(xs, slow_fcn) |> progressify()
```

Using the default progress handler, the progress reporting will
appear as:

```plain
  |=====                    |  20%
```


# Supported Functions

The `progressify()` function supports the following **base** package
functions:

 * `lapply()`, `vapply()`, `sapply()`, `tapply()`
 * `mapply()`, `.mapply()`, `Map()`
 * `eapply()`
 * `apply()`
 * `replicate()`
 * `by()`


# Combining with futurize

The **progressify** package works together with the **[futurize]**
package. You can both parallelize and add progress reporting in a
single pipeline:

```r
library(futurize)
plan(multisession)
library(progressify)
handlers(global = TRUE)

xs <- 1:100
ys <- lapply(xs, slow_fcn) |> futurize() |> progressify()
```


# Known issues

The **[BiocGenerics]** package defines generic functions `lapply()`,
`sapply()`, `mapply()`, and `tapply()`. These S4 generic functions
override the non-generic, counterpart functions in the **base**
package. If **BiocGenerics** is attached, the solution is to specify
that it is the **base** version we wish to progressify, i.e.

```r
y <- base::lapply(1:3, sqrt) |> progressify()
```


[futurize]: https://cran.r-project.org/package=futurize
[BiocGenerics]: https://www.bioconductor.org/packages/BiocGenerics/

<!--
%\VignetteIndexEntry{Progress updates for 'boot' functions}
%\VignetteAuthor{Henrik Bengtsson}
%\VignetteKeyword{R}
%\VignetteKeyword{package}
%\VignetteKeyword{boot}
%\VignetteKeyword{vignette}
%\VignetteKeyword{progressify}
%\VignetteEngine{progressify::selfonly}
-->

The **progressify** package allows you to easily add progress
reporting to sequential and parallel map-reduce code by piping to the
`progressify()` function. Easy!


# TL;DR

```r
library(progressify)
handlers(global = TRUE)
library(boot)

# Run bootstrap with progress signaling
x <- 1:100
my_stat <- function(data, i) mean(data[i])
res <- boot(data = x, statistic = my_stat, R = 1000) |> progressify()
```


# Introduction

This vignette demonstrates how to use this approach to add progress
reporting to **[boot]** functions such as `boot()`, `censboot()`, and `tsboot()`.

The **boot** package provides functions for generating bootstrap replicates.
Because these computations are iterative and computationally intensive, they can
benefit significantly from progress reporting.

For example, `boot()` runs a statistic function `R` times:

```r
library(boot)
x <- 1:100
my_stat <- function(data, i) mean(data[i])
res <- boot(data = x, statistic = my_stat, R = 1000)
```

By default, `boot()` provides no feedback on how far it has progressed.
However, we can easily add progress reporting using the `progressify()` function:

```r
library(boot)

library(progressify)
handlers(global = TRUE)

x <- 1:100
my_stat <- function(data, i) mean(data[i])
res <- boot(data = x, statistic = my_stat, R = 1000) |> progressify()
```


# Supported Functions

The `progressify()` function supports the following **boot**
functions:

 * `boot()`
 * `censboot()`
 * `tsboot()`


[boot]: https://cran.r-project.org/package=boot

<!--
%\VignetteIndexEntry{Progress updates for 'foreach' functions}
%\VignetteAuthor{Henrik Bengtsson}
%\VignetteKeyword{R}
%\VignetteKeyword{package}
%\VignetteKeyword{foreach}
%\VignetteKeyword{doFuture}
%\VignetteKeyword{vignette}
%\VignetteKeyword{progressify}
%\VignetteEngine{progressify::selfonly}
-->

The **progressify** package allows you to easily add progress
reporting to sequential and parallel map-reduce code by piping to the
`progressify()` function. Easy!


# TL;DR

```r
library(progressify)
handlers(global = TRUE)
library(foreach)

slow_fcn <- function(x) {
  Sys.sleep(0.1)  # emulate work
  x^2
}

xs <- 1:100
ys <- foreach(x = xs) %do% slow_fcn(x) |> progressify()
```


# Introduction

This vignette demonstrates how to use this approach to add progress
reporting to the **[foreach]** `foreach()` construct and the
**[doFuture]** `%dofuture%` operator.

For example, consider:

```r
library(foreach)
xs <- 1:100
ys <- foreach(x = xs) %do% slow_fcn(x)
```

This `foreach()` construct provides no feedback on how far it has
progressed. We can easily add progress reporting by piping to
`progressify()`:

```r
library(foreach)

library(progressify)
handlers(global = TRUE)

xs <- 1:100
ys <- foreach(x = xs) %do% slow_fcn(x) |> progressify()
```

Using the default progress handler, the progress reporting will
appear as:

```plain
  |=====                    |  20%
```


## With doFuture

The same approach works with the **[doFuture]** package for parallel
foreach evaluation:

```r
library(doFuture)
plan(multisession)

library(progressify)
handlers(global = TRUE)

xs <- 1:100
ys <- foreach(x = xs) %dofuture% slow_fcn(x) |> progressify()
```


# Supported Functions

The `progressify()` function supports the following **foreach**
operators:

 * `foreach(...) %do% { ... }`
 * `foreach(...) %dopar% { ... }`

and the following **doFuture** operator:

 * `foreach(...) %dofuture% { ... }`


[foreach]: https://cran.r-project.org/package=foreach
[doFuture]: https://cran.r-project.org/package=doFuture

<!--
%\VignetteIndexEntry{Progress updates for 'crossmap' functions}
%\VignetteAuthor{Henrik Bengtsson}
%\VignetteKeyword{R}
%\VignetteKeyword{package}
%\VignetteKeyword{crossmap}
%\VignetteKeyword{vignette}
%\VignetteKeyword{progressify}
%\VignetteEngine{progressify::selfonly}
-->

The **progressify** package allows you to easily add progress
reporting to sequential and parallel map-reduce code by piping to the
`progressify()` function. Easy!


# TL;DR

```r
library(progressify)
handlers(global = TRUE)
library(crossmap)

slow_fcn <- function(x, y) {
  Sys.sleep(0.1)  # emulate work
  x * y
}

xs <- list(1:5, 1:5)
ys <- xmap(xs, slow_fcn) |> progressify()
```


# Introduction

This vignette demonstrates how to use this approach to add progress
reporting to **[crossmap]** functions such as `xmap()`.

The **crossmap** package extends **purrr** with functions that apply
a function to every combination of elements in a list of inputs.
For example, `xmap()` computes the cross product of its inputs:

```r
library(crossmap)
xs <- list(1:5, 1:5)
ys <- xmap(xs, slow_fcn)
```

Here `xmap()` provides no feedback on how far it has progressed,
but we can easily add progress reporting by using:

```r
library(crossmap)

library(progressify)
handlers(global = TRUE)

xs <- list(1:5, 1:5)
ys <- xmap(xs, slow_fcn) |> progressify()
```

Using the default progress handler, the progress reporting will
appear as:

```plain
  |=====                    |  20%
```


# Supported Functions

The `progressify()` function supports the following **crossmap**
functions:

 * `xmap()` and variants (`xmap_chr()`, `xmap_dbl()`, `xmap_int()`,
   `xmap_lgl()`, `xmap_vec()`, `xmap_dfc()`, `xmap_dfr()`,
   `xmap_mat()`, `xmap_arr()`)
 * `xwalk()`
 * `map_vec()`, `map2_vec()`, `pmap_vec()`, `imap_vec()`
 * `future_xmap()` and variants (`future_xmap_chr()`,
   `future_xmap_dbl()`, `future_xmap_int()`, `future_xmap_lgl()`,
   `future_xmap_vec()`, `future_xmap_dfc()`, `future_xmap_dfr()`,
   `future_xmap_mat()`, `future_xmap_arr()`, `future_xmap_raw()`)
 * `future_xwalk()`
 * `future_map_vec()`, `future_map2_vec()`, `future_pmap_vec()`,
   `future_imap_vec()`


[crossmap]: https://cran.r-project.org/package=crossmap

<!--
%\VignetteIndexEntry{Progress updates for 'partykit' functions}
%\VignetteAuthor{Henrik Bengtsson}
%\VignetteKeyword{R}
%\VignetteKeyword{package}
%\VignetteKeyword{partykit}
%\VignetteKeyword{vignette}
%\VignetteKeyword{progressify}
%\VignetteEngine{progressify::selfonly}
-->

The **progressify** package allows you to easily add progress
reporting to sequential and parallel map-reduce code by piping to the
`progressify()` function. Easy!


# TL;DR

```r
library(progressify)
handlers(global = TRUE)
library(partykit)

data("Titanic", package = "datasets")
tt <- as.data.frame(Titanic)

forest <- cforest(Survived ~ ., data = tt, ntree = 50L) |> progressify()
```


# Introduction

This vignette demonstrates how to use this approach to add progress
reporting to **[partykit]** functions such as `cforest()`.

The **partykit** `cforest()` function is an implementation of random
forests. For example,

```r
library(partykit)
data("Titanic", package = "datasets")
tt <- as.data.frame(Titanic)
forest <- cforest(Survived ~ ., data = tt, ntree = 50L)
```

Here `cforest()` provides no feedback on how far it has progressed,
but we can easily add progress reporting by using:

```r
library(partykit)

library(progressify)
handlers(global = TRUE)

data("Titanic", package = "datasets")
tt <- as.data.frame(Titanic)

forest <- cforest(Survived ~ ., data = tt, ntree = 50L) |> progressify()
```

Using the default progress handler, the progress reporting will
appear as:

```plain
  |=====                    |  20%
```


# Supported Functions

The `progressify()` function supports the following **partykit**
functions:

 * `cforest()`


[partykit]: https://cran.r-project.org/package=partykit

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/progressify.R
\name{progressify}
\alias{progressify}
\alias{pz}
\title{Evaluate a regular map-reduce call with progress updates}
\usage{
progressify(
  expr,
  substitute = TRUE,
  ...,
  when = TRUE,
  eval = TRUE,
  envir = parent.frame()
)
}
\arguments{
\item{expr}{An R expression.}

\item{substitute}{If TRUE, \code{expr} is quoted.}

\item{when}{If TRUE (default), the expression is progressified, otherwise not.}

\item{eval}{If TRUE (default), the progressified expression is evaluated,
otherwise it is returned.}

\item{envir}{The environment in which \code{expr} is evaluated.}

\item{\ldots}{Not used.}
}
\value{
Returns the value of the evaluated expression \code{expr}.
}
\description{
Evaluate a regular map-reduce call with progress updates
}
\examples{
handlers(global = TRUE) # listen to progress updates

xs <- list(1, 1:2, 1:2, 1:5)
y <- lapply(X = xs, FUN = sum) |> progressify()
str(y)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/transpile.R
\name{transpile}
\alias{transpile}
\title{Transpile an R expression}
\usage{
transpile(
  expr,
  options = list(...),
  ...,
  when = TRUE,
  eval = TRUE,
  envir = parent.frame(),
  disable = FALSE,
  type = "built-in",
  what = "transpile",
  unwrap = list(base::`{`, base::`(`, base::`!`, base::local, base::with, base::I,
    base::identity, base::invisible, base::suppressMessages, base::suppressWarnings,
    base::suppressPackageStartupMessages),
  debug = FALSE
)
}
\arguments{
\item{expr}{An \R expression, typically a function call to transpile.
If FALSE, then the transpiler is disabled, and if TRUE, it is re-enabled.
If NA, then TRUE is returned if the transpiler is enabled, otherwise FALSE.}

\item{options}{(optional) Named options for the transpilation.}

\item{when}{If TRUE (default), the expression is transpiled, otherwise not.}

\item{eval}{If TRUE (default), the transpiled expression is evaluated,
otherwise it is returned.}

\item{envir}{The environment where the expression should be evaluated.}

\item{type}{Type of the transpiler to use.}

\item{unwrap}{(optional) A list of functions that should be considered
wrapping functions that the transpiler should unwrap ("enter"). This
allows us to transpile expressions within \code{{ ... }} and \code{local( ... )}.}
}
\value{
Returns the value of the evaluated expression \code{expr} if \code{eval = TRUE},
otherwise the transpiled expression.
If \code{expr} is NA, then TRUE is returned if the transpiler is enabled,
otherwise FALSE.
}
\description{
Transpile an R expression
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/progressify_supported_tools.R
\name{progressify_supported_packages}
\alias{progressify_supported_packages}
\alias{progressify_supported_functions}
\title{List packages and functions supporting progressification}
\usage{
progressify_supported_packages()

progressify_supported_functions(package)
}
\arguments{
\item{package}{A package name.}
}
\value{
A character vector of package or function names.
}
\description{
List packages and functions supporting progressification
}
\examples{
pkgs <- progressify_supported_packages()
pkgs

fcns <- progressify_supported_functions("base")
print(fcns)

if (requireNamespace("purrr")) {
  fcns <- progressify_supported_functions("purrr")
  print(fcns)
}

}
